import doctest
import warnings

from nglview import show, viewer_control, widget

warnings.filterwarnings('ignore')


def get_total_errors(modules):
    return sum([doctest.testmod(mod).failed for mod in modules])


def test_nglview_show_module():
    """
    """
    assert not get_total_errors([show])


def test_nglview_viewer_control():
    """
    """
    assert not get_total_errors([viewer_control])


def test_nglview_widget():
    """
    """
    assert not get_total_errors([widget])



import time

from ..parameters import REPRESENTATION_NAME_PAIRS


def wait(widget, attribute='value', timeout=5):
    """ EXPERIMENTAL. Require `ipython_blocking` package.

    Block further code execution until `attribute` of `widget` is updated.
    """
    from ipython_blocking import CaptureExecution
    c = CaptureExecution()
    with c:
        t0 = time.time()
        while True or time.time() - t0 > timeout:
            attr = getattr(widget, attribute)
            if attr:
                break
            c.step()
    return widget


def get_widget_by_name(box, widget_name):

    if hasattr(box, '_ngl_children'):
        children = box._ngl_children
    elif hasattr(box, 'children'):
        children = box.children
    else:
        children = None

    if children is not None:
        for widget in children:
            if hasattr(widget,
                       '_ngl_name') and widget._ngl_name == widget_name:
                return widget
    return None


def _add_repr_method_shortcut(self, other):
    from types import MethodType

    def make_func_add(rep):
        """return a new function object
        """

        def func(this, selection='all', **kwargs):
            """
            """
            self.add_representation(repr_type=rep[1],
                                    selection=selection,
                                    **kwargs)

        func.__doc__ = """Shortcut for `add_representation` method

        Examples
        --------
        >>> view.add_{name}()
        >>> # is equal to
        >>> view.add_representation('{name}')
        """.format(name=rep[0])
        return func

    def make_func_remove(rep):
        """return a new function object
        """

        def func(this, **kwargs):
            """
            """
            self._remove_representations_by_name(repr_name=rep[1], **kwargs)

        return func

    def make_func_update(rep):
        """return a new function object
        """

        def func(this, **kwargs):
            """
            """
            self._update_representations_by_name(repr_name=rep[1], **kwargs)

        return func

    for rep in REPRESENTATION_NAME_PAIRS:
        for make_func, root_fn in [(make_func_add, 'add'),
                                   (make_func_update, 'update'),
                                   (make_func_remove, 'remove')]:
            func = make_func(rep)
            fn = '_'.join((root_fn, rep[0]))
            setattr(self, fn, MethodType(func, other))



from ipywidgets import DOMWidget
from traitlets import Bool, List, observe


def _singleton(cls):
    # https://www.python.org/dev/peps/pep-0318/#examples
    instances = {}
    def getinstance():
        if cls not in instances:
            instances[cls] = cls()
        return instances[cls]
    return getinstance


class BaseWidget(DOMWidget):
    _msg_q = []
    _msg_ar = List().tag(sync=True)
    _ready = Bool(False).tag(sync=True)

    def _js(self, code):
        self._call("executeCode", code)

    def _call(self, method_name, *args, **kwargs):
        msg = {"type": "callMethod",
               "methodName": method_name,
                "args": args, "kwargs": kwargs}
        if not self._ready:
            # fire later
            self._msg_q.append(msg)
        else:
            self.send(msg)
        msg_ar = self._msg_ar[:]
        msg_ar.append(msg)
        self._msg_ar = msg_ar # trigger sync

    @observe("_ready")
    def _on_ready(self, change):
        if change.new:
            while self._msg_q:
                msg = self._msg_q.pop(0)
                self.send(msg)

    def _ipython_display_(self, **kwargs):
        try:
            # ipywidgets < 8
            super()._ipython_display_(**kwargs)
        except AttributeError:
            display(super()._repr_mimebundle_(), raw=True)



import os

from mock import MagicMock, patch

import nglview
from make_dummy_comm import *
import PIL.Image


@patch('moviepy.editor.ImageSequenceClip')
@patch('PIL.Image')
def test_movie_maker(mock_image, ImageSequenceClip):
    from nglview.contrib.movie import MovieMaker
    ImageSequenceClip.write_gif = MagicMock()
    ImageSequenceClip.write_videofile = MagicMock()
    traj = nglview.SimpletrajTrajectory(nglview.datafiles.XTC, nglview.datafiles.PDB)
    view = nglview.show_simpletraj(traj)

    movie = MovieMaker(view,
                       in_memory=True,
                       download_folder='here',
                       render_params={'factor': 4},
                       moviepy_params={},
                       stop=2)
    movie.make()



from typing import List
try:
    import moviepy.editor as mpy
except ImportError:
    print("You have to install moviepy, imageio and ffmeg")
    print("pip install moviepy==0.2.2.11")
    print("pip install imageio==1.6")

import os
import threading
import time
from ipywidgets import Button, Output, IntProgress
from itertools import tee


class MovieMaker:
    """ Unstable API

    Parameters
    ----------
    view : NGLWidget
    download_folder : str or None
        Folder that stores images. You can not arbitarily set this folder. It must be
        the download directory of the web browser you are using.
        If None, $HOME/Downloads/ will be used.
        NOTE: This is DEPRECATED (used with `make_old_impl`)
    prefix : str, default 'movie'
        prefix name of rendered image.
    output : str, default 'my_movie.gif'
        output filename of the movie.
        if output has '.gif', call `write_gif`, otherwise calling `write_videofile`
    fps : 8
        frame per second
    start, stop, step : int, default (0, -1, 1)
        how many frames you want to render.
    skip_render : bool, default False
        if True, do not render any frame and uses existings images in `download_folder`
        for movie making.
        if False, perform rendering first.
    timeout : a number (second), default 0.1
        The waiting time between rendering two consecutive frames.
        This option should be only used with `perframe_hook` option.
    render_params : dict or None, default None
        NGL rendering params. see NGLWidget.download_image.
        If None, use default values
    moviepy_params : dict or None, default None
        moviepy params for `write_gif` method.
        if None, use default values
    in_memory : bool, default False
        if False, save rendered images to disk first
        if True, keep all image data in memory (good for small video)
    perframe_hook : callable with `view` as a single argument, default None
        if given, update the `view` by `perframe_hook`.

    Examples
    --------
    >>> import nglview as nv
    >>> import pytraj as pt
    >>> traj = pt.load(nv.datafiles.XTC, top=nv.datafiles.PDB)
    >>> view = nv.show_pytraj(traj)
    >>> from nglview.contrib.movie import MovieMaker
    >>> download_folder = '/Users/xxx/Downloads'
    >>> output = 'my.gif'
    >>> mov = MovieMaker(view, download_folder=download_folder, output=output)
    >>> mov.make()

    >>> # write avi format
    >>> from nglview.contrib.movie import MovieMaker

    >>> moviepy_params = {
    ...     'codec': 'mpeg4'
    ... }
    >>> movie = MovieMaker(view, output='my.avi', in_memory=True, moviepy_params=moviepy_params)
    >>> movie.make()

    Notes
    -----
    unstable API. Currently supports .gif format
    If you are using remote notebook, make sure to set in_memory=True

    Requires
    --------

    moviepy. e.g:

        pip install moviepy
        conda install freeimage

    """

    def __init__(self,
                 view,
                 download_folder=None,
                 prefix='movie',
                 output='my_movie.gif',
                 fps=8,
                 start=0,
                 stop=-1,
                 step=1,
                 skip_render=False,
                 timeout=0.1,
                 in_memory=False,
                 perframe_hook=None,
                 render_params=None,
                 moviepy_params=None):
        if download_folder is None:
            download_folder = os.getenv('HOME', '') + '/Downloads/'
        self.view = view
        self.skip_render = skip_render
        self.prefix = prefix
        self.download_folder = download_folder
        self.timeout = timeout
        self.fps = fps
        self.in_memory = in_memory
        self.render_params = render_params or dict(
            factor=4, antialias=True, trim=False, transparent=False)
        self.moviepy_params = moviepy_params or {}
        self.perframe_hook = perframe_hook
        self.output = output
        if stop < 0:
            stop = self.view.max_frame + 1
        self._time_range = range(start, stop, step)
        self._iframe = iter(self._time_range)
        self._progress = IntProgress(max=len(self._time_range) - 1)
        self._woutput = Output()
        self._event = threading.Event()
        self._thread = None
        self._image_array = []

    def sleep(self):
        time.sleep(self.timeout)

    def make_old_impl(self, in_memory=False):
        # TODO : make base class so we can reuse this with sandbox/base.py
        progress = IntProgress(
            description='Rendering...', max=len(self._time_range) - 1)
        self._event = threading.Event()

        def _make(event):
            image_files = []
            iw = None
            if not self.skip_render:
                for i in self._time_range:
                    progress.value = i
                    if not event.is_set():
                        self.view.frame = i
                        self.sleep()
                        if self.perframe_hook:
                            self.perframe_hook(self.view)
                        self.sleep()
                        if not self.in_memory:
                            self.view.download_image(
                                self.prefix + '.' + str(i) + '.png',
                                **self.render_params)
                        else:
                            iw = self.view.render_image(**self.render_params)
                        self.sleep()
                        if self.in_memory:
                            rgb = self._base64_to_ndarray(
                                self.view._image_data)
                            self._image_array.append(rgb)
                            if iw:
                                iw.close()  # free memory
                if not self.in_memory:
                    template = "{}/{}.{}.png"
                    image_files = [
                        image_dir
                        for image_dir in (template.format(
                            self.download_folder, self.prefix, str(i))
                                          for i in self._time_range)
                        if os.path.exists(image_dir)
                    ]
                else:
                    image_files = self._image_array
            if not self._event.is_set():
                progress.description = "Writing ..."
                clip = mpy.ImageSequenceClip(image_files, fps=self.fps)
                with Output():
                    if self.output.endswith('.gif'):
                        clip.write_gif(
                            self.output,
                            fps=self.fps,
                            verbose=False,
                            **self.moviepy_params)
                    else:
                        clip.write_videofile(
                            self.output, fps=self.fps, **self.moviepy_params)
                self._image_array = []
                progress.description = 'Done'
                time.sleep(1)
                progress.close()

        self.thread = threading.Thread(target=_make, args=(self._event, ))
        self.thread.daemon = True
        self.thread.start()
        return progress

    def make(self, movie=True, keep_data=False):
        """

        Parameters
        ----------
        keep_data: bool
            if True, save the image data in self._image_array
        movie: bool
            if True, make the movie
            else, only do the rendering (make sure keep_data=True in this case)
        """
        self._woutput.clear_output()
        image_array = []
        iframe = iter(self._time_range)
        frame = next(iframe)

        def hook(frame):
            self.view.frame = frame
            time.sleep(self.timeout)
            self.perframe_hook(self.view)
            time.sleep(self.timeout)

        # trigger movie making communication between backend and frontend
        self.perframe_hook and hook(frame)
        self.view._set_coordinates(
            frame, movie_making=True, render_params=self.render_params)
        self._progress.description = 'Rendering ...'

        def on_msg(widget, msg, _):
            if msg['type'] == 'movie_image_data':
                image_array.append(msg.get('data'))
                try:
                    frame = next(iframe)
                    self.perframe_hook and hook(frame)
                    self.view._set_coordinates(
                        frame,
                        movie_making=True,
                        render_params=self.render_params)
                    self._progress.value = frame
                except StopIteration:
                    if movie:
                        self._progress.description = 'Making movie...'
                        with self._woutput:
                            # suppress moviepy's log
                            self._make_from_array(image_array)
                        if not os.path.exists(self.output):
                            self._progress.description = "ERROR: Check the maker's log"
                        else:
                            self._progress.description = 'Done'
                    self._remove_on_msg()
                    if keep_data:
                        self._image_array = image_array

        self._on_msg = on_msg
        # FIXME: if exception happens, the on_msg callback will be never removed
        # from `self.view`
        self.view.on_msg(on_msg)
        return self._progress

    def _remove_on_msg(self):
        self.view.on_msg(self._on_msg, remove=True)

    def _make_from_array(self, image_array: List[str]):
        image_files = [self._base64_to_ndarray(a) for a in image_array]
        clip = mpy.ImageSequenceClip(image_files, fps=self.fps)
        with self._woutput:
            if self.output.endswith('.gif'):
                clip.write_gif(
                    self.output,
                    fps=self.fps,
                    verbose=False,
                    **self.moviepy_params)
            else:

                clip.write_videofile(
                    self.output, fps=self.fps, **self.moviepy_params)

    def interupt(self):
        """ Stop making process """
        if self._event is not None:
            self._event.set()

    @classmethod
    def _base64_to_ndarray(cls, value):
        import io
        import base64
        from PIL import Image
        import numpy as np
        im_bytes = base64.b64decode(value)
        im_bytes = io.BytesIO(im_bytes)
        # convert to numpy RGB value (for moviepy.editor.VideoClip)
        return np.array(Image.open(im_bytes))




import nglview as nv
from nglview.theme import ThemeManager


def test_theme():
    m  = ThemeManager()
    assert m is ThemeManager() # singleton
    assert m._theme_css == ''
    m.dark()
    assert m._theme == 'dark'
    assert m._theme_css is not None
    m.light()
    assert m._theme == 'light'
    m.remove()
    assert m._theme is None
    assert m._theme_css == ''



from .utils.py_utils import _camelize_dict


class Stage:
    """
    Try to mimic NGL.Stage
    """

    def __init__(self, view):
        self._view = view

    def set_parameters(self, **kwargs):
        self._view._remote_call('setParameters',
                                target='Stage',
                                kwargs=_camelize_dict(kwargs))



import os
import subprocess
import sys
import unittest

import pytest
from mock import mock_open, patch

from nglview import datafiles
from nglview.scripts.nglview import get_remote_port, main

this_path = os.path.dirname(os.path.abspath(__file__))


@patch('subprocess.check_call')
def test_cli(mock_call):
    # no argument
    command = []
    # subprocess.check_call(command.split())
    main(cmd=command)

    # only nglview and --clean
    command = ['--clean']
    main(cmd=command)

    # demo
    command = ['demo']
    main(cmd=command)

    # single pdb
    command = [datafiles.PDB]
    main(cmd=command)

    # single pdb, disable-autorun
    command = [datafiles.PDB, '--auto']
    main(cmd=command)

    # single pdb, specify browser
    command = f'nglview {datafiles.PDB} --browser=google-chrome'
    subprocess.check_call(command.split())

    # single pdb, does not exists
    command = ['test_hellooooo.pdb']
    with pytest.raises(AssertionError):
        main(cmd=command)

    # density data
    densityfile = os.path.join(this_path, 'data/volmap.dx')
    command = [densityfile]
    main(cmd=command)

    # python script
    command = ['my.py']
    with patch('builtins.open'), patch('json.dumps'):
        main(cmd=command)

    # pytraj
    command = f'nglview {datafiles.PDB} -c {datafiles.XTC}'
    subprocess.check_call(command.split())

    # remote
    command = 'nglview {} -c {} --remote'.format(datafiles.PDB, datafiles.XTC)
    subprocess.check_call(command.split())

    # python script
    pyfile = os.path.join(this_path, 'test_widget.py')
    command = f'nglview {pyfile}'
    subprocess.check_call(command.split())

    # notebook (.ipynb)
    nbfile = os.path.join(this_path, 'notebooks/api/test_detach.ipynb')
    command = [nbfile]
    main(cmd=command)

    # install
    command = ['install', '--symlink']
    with pytest.raises(SystemExit):
        main(cmd=command)
    mock_call.assert_called_with([
        'jupyter', 'nbextension', 'install', '--py', '--sys-prefix', 'nglview',
        '--overwrite', '--symlink'
    ])

    # enable
    command = ['enable']
    with pytest.raises(SystemExit):
        main(cmd=command)
    mock_call.assert_called_with([
        'jupyter', 'nbextension', 'enable', '--py', '--sys-prefix', 'nglview'
    ])

    # remote
    with patch('nglview.scripts.nglview.get_remote_port') as mock_remote:
        command = ['--remote']
        main(cmd=command)
        mock_remote.assert_called_with(None, 'tmpnb_ngl.ipynb')


@patch('nglview.scripts.app.NGLViewApp.get_port')
@patch('os.getlogin')
@patch('socket.gethostname')
def test_get_port(mock_gethostname, mock_getlogin, mock_get_port):
    mock_gethostname.return_value = 'myhost'
    mock_getlogin.return_value = 'mylogin'
    mock_get_port.return_value = 9999
    get_remote_port(None, 'my.ipynb')



"""This module is meant for playground.
"""

import numpy as np


def smooth(coords_3d,
           method='filtfilt',
           inplace=True,
           atom_indices=None,
           **kwargs):
    """
    Parameters
    ----------
    coords_3d : 3d numpy's array
    method : str, {'filtfilt', 'savgol_filter'}
    atom_indices : None or 1d integer array-like
        If None, using all atoms.
    kwargs : Additional arguments for corresponding method.

    Examples
    --------
    >>> import nglview as nv
    >>> from nglview.sandbox.interpolate import smooth
    >>> import pytraj as pt
    >>> traj = pt.datafiles.load_tz2()[:]
    >>> traj.xyz = smooth(traj.xyz)
    >>> view = nv.show_pytraj(traj)
    >>> view.clear()
    >>> view.add_licorice('protein')
    >>> view
    """
    from scipy import signal
    if inplace:
        xyz = coords_3d
    else:
        xyz = coords_3d.copy()
    n_atoms = coords_3d[0].shape[0]
    xyz0 = xyz[0].copy()
    atom_indices = atom_indices or range(n_atoms)
    for idx in atom_indices:
        for j in range(3):
            xyz_idx_j = xyz[:, idx, j]
            if method == 'filtfilt':
                b, a = signal.butter(3, 0.1)
                smooth_data = signal.filtfilt(b, a, xyz_idx_j, **kwargs)
            elif method == 'savgol_filter':
                smooth_data = signal.savgol_filter(xyz_idx_j, **kwargs)
            else:
                raise ValueError(f"%s method is not supported" % method)
            xyz[:, idx, j] = smooth_data
    xyz[0] = xyz0  # to correctly render 1st frame.
    return xyz



import warnings

# for doc
from . import adaptor, datafiles, show, widget
from ._version import get_versions
from .adaptor import *
from .base_adaptor import *
# TODO: do not use import *
# interface
from .config import BACKENDS
from .data_source import DatasourceRegistry
from .show import *
# utils
from .utils import js_utils, widget_utils
from .widget import NGLWidget, write_html

__version__ = get_versions()['version']
del get_versions

with warnings.catch_warnings():
    warnings.simplefilter("ignore")

# Register nbextension


def _jupyter_nbextension_paths():
    return [{
        'section': 'notebook',
        'src': 'static',
        'dest': 'nglview-js-widgets',
        'require': 'nglview-js-widgets/extension'
    }]


__all__ = ['NGLWidget', 'write_html'
           ] + widget.__all__ + adaptor.__all__ + show.__all__



#-----------------------------------------------------------------------------
# Utility stuff from ipywidgets tests
# https://github.com/jupyter-widgets/ipywidgets/blob/8.0.5/python/ipywidgets/ipywidgets/widgets/tests/utils.py
#
# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.

#-----------------------------------------------------------------------------

from ipywidgets import Widget
import ipywidgets.widgets.widget

# The new comm package is not available in our Python 3.7 CI (older ipykernel version)
try:
    import comm
    NEW_COMM_PACKAGE = True
except ImportError:
    NEW_COMM_PACKAGE = False

import ipykernel.comm


class DummyComm():
    comm_id = 'a-b-c-d'
    kernel = 'Truthy'

    def __init__(self, *args, **kwargs):
        super().__init__()
        self.messages = []

    def open(self, *args, **kwargs):
        pass

    def on_msg(self, *args, **kwargs):
        pass

    def send(self, *args, **kwargs):
        self.messages.append((args, kwargs))

    def close(self, *args, **kwargs):
        pass


def dummy_create_comm(**kwargs):
    return DummyComm()


def dummy_get_comm_manager(**kwargs):
    return {}


_widget_attrs = {}
undefined = object()

if NEW_COMM_PACKAGE:
    orig_comm = ipykernel.comm.comm.BaseComm
else:
    orig_comm = ipykernel.comm.Comm
orig_create_comm = None
orig_get_comm_manager = None

if NEW_COMM_PACKAGE:
    orig_create_comm = comm.create_comm
    orig_get_comm_manager = comm.get_comm_manager

def setup_test_comm():
    if NEW_COMM_PACKAGE:
        comm.create_comm = dummy_create_comm
        comm.get_comm_manager = dummy_get_comm_manager
        ipykernel.comm.comm.BaseComm = DummyComm
    else:
        ipykernel.comm.Comm = DummyComm
    Widget.comm.klass = DummyComm
    ipywidgets.widgets.widget.Comm = DummyComm
    _widget_attrs['_repr_mimebundle_'] = Widget._repr_mimebundle_
    def raise_not_implemented(*args, **kwargs):
        raise NotImplementedError()
    Widget._repr_mimebundle_ = raise_not_implemented

def teardown_test_comm():
    if NEW_COMM_PACKAGE:
        comm.create_comm = orig_create_comm
        comm.get_comm_manager = orig_get_comm_manager
        ipykernel.comm.comm.BaseComm = orig_comm
    else:
        ipykernel.comm.Comm = orig_comm
    Widget.comm.klass = orig_comm
    ipywidgets.widgets.widget.Comm = orig_comm
    for attr, value in _widget_attrs.items():
        if value is undefined:
            delattr(Widget, attr)
        else:
            setattr(Widget, attr, value)
    _widget_attrs.clear()

def setup():
    setup_test_comm()

def teardown():
    teardown_test_comm()

def call_method(method, *args, **kwargs):
    method(*args, **kwargs)



import base64
import json
import threading
import time
import uuid
from logging import getLogger

import ipywidgets as widgets
import ipywidgets.embed
import numpy as np
from IPython.display import display
from ipywidgets import (Image, Box, DOMWidget, HBox, VBox, IntSlider, Output, Play, Widget,
                        jslink)
from ipywidgets import widget as _widget
from traitlets import (Bool, CaselessStrEnum, Dict, Instance, Int, Integer,
                       List, Unicode, observe, validate)
import traitlets

from . import color, interpolate
from .adaptor import Structure, Trajectory
from .component import ComponentViewer
from .config import BACKENDS
from .player import TrajectoryPlayer, _dry_run
from .remote_thread import RemoteCallThread
from .representation import RepresentationControl
from .shape import Shape
from .stage import Stage
from .utils import py_utils, widget_utils
from .utils.py_utils import (FileManager, _camelize_dict, _update_url,
                             encode_base64, get_repr_names_from_dict,
                             seq_to_string)
from .viewer_control import ViewerControl
from ._frontend import __frontend_version__
from .base import BaseWidget

logger = getLogger(__name__)

widget_serialization = _widget.widget_serialization

__all__ = ['NGLWidget', 'ComponentViewer']
_EXCLUDED_CALLBACK_AFTER_FIRING = {
    'setUnSyncCamera',
    'setSelector',
    'setDelay',
    'autoView',
    '_downloadImage',
    '_exportImage',
    'set_representation_from_backend',
}

_INIT_VIEWS = {'color_maker_registry': color.ColormakerRegistry}
_TRACKED_WIDGETS = {}


def _deprecated(msg):
    def wrap_1(func):
        def wrap_2(*args, **kwargs):
            logger.warn(msg)
            return func(*args, **kwargs)

        return wrap_2

    return wrap_1


def write_html(fp, views, frame_range=None):
    # type: (str, List[NGLWidget]) -> None
    """EXPERIMENTAL. Likely will be changed.

    Make html file to display a list of views. For further options, please
    check `ipywidgets.embed` module.

    Parameters
    ----------
    fp : str or file handle
    views : a DOMWidget view or a list of views.
    frame_range : None or a tuple of int

    Examples
    --------
    >>> import nglview
    >>> view = nglview.show_pdbid('1tsu')
    >>> view # doctest: +SKIP
    >>> nglview.write_html('index.html', [view]) # doctest: +SKIP
    >>> nglview.write_html('index.html', [view], frame_range=(0, 5)) # doctest: +SKIP
    """
    views = isinstance(views, DOMWidget) and [views] or views
    embed = ipywidgets.embed
    color = None
    theme = None

    for _, v in _INIT_VIEWS.items():
        views.insert(0, v)

    def _set_serialization(views):
        for view in views:
            if hasattr(view, '_set_serialization'):
                view._set_serialization(frame_range=frame_range)
            elif isinstance(view, Box):
                _set_serialization(view.children)

    def _unset_serialization(views):
        for view in views:
            if hasattr(view, '_unset_serialization'):
                view._unset_serialization()
            elif isinstance(view, Box):
                _unset_serialization(view.children)

    _set_serialization(views)
    # FIXME: allow add jquery-ui link?
    snippet = '<link rel="stylesheet" href="https://cdnjs.cloudflare.com/ajax/libs/jqueryui/1.12.0/jquery-ui.css">\n'
    snippet += embed.embed_snippet(views)
    html_code = embed.html_template.format(title='nglview-demo',
                                           snippet=snippet)

    # from ipywidgets
    # Check if fp is writable:
    if hasattr(fp, 'write'):
        fp.write(html_code)
    else:
        # Assume fp is a filename:
        with open(fp, "w") as f:
            f.write(html_code)

    _unset_serialization(views)



class NGLWidget(DOMWidget):
    _view_name = Unicode("NGLView").tag(sync=True)
    _view_module = Unicode("nglview-js-widgets").tag(sync=True)
    _view_module_version = Unicode(__frontend_version__).tag(sync=True)
    _model_name = Unicode("NGLModel").tag(sync=True)
    _model_module = Unicode("nglview-js-widgets").tag(sync=True)
    _model_module_version = Unicode(__frontend_version__).tag(sync=True)
    _ngl_version = Unicode().tag(sync=True)
    # _model_name = Unicode("NGLView").tag(sync=True)
    # _model_module = Unicode("nglview-js-widgets").tag(sync=True)
    _image_data = Unicode().tag(sync=False)
    # use Integer here, because mdtraj uses a long datatype here on Python-2.7
    frame = Integer().tag(sync=True)
    max_frame = Int(0).tag(sync=True)
    background = Unicode('white').tag(sync=True)
    loaded = Bool(False).tag(sync=False)
    picked = Dict().tag(sync=True)
    n_components = Int(0).tag(sync=True)
    _view_width = Unicode().tag(sync=True) # px
    _view_height = Unicode().tag(sync=True) # px
    _scene_position = Dict().tag(sync=True)
    _scene_rotation = Dict().tag(sync=True)
    # hack to always display movie
    # TODO: remove _parameters?
    _parameters = Dict().tag(sync=False)
    _ngl_full_stage_parameters = Dict().tag(sync=True)
    _ngl_original_stage_parameters = Dict().tag(sync=True)
    _coordinates_dict = Dict().tag(sync=False)
    _camera_str = CaselessStrEnum(['perspective', 'orthographic'],
                                  default_value='orthographic').tag(sync=True)
    _camera_orientation = List().tag(sync=True)
    _synced_model_ids = List().tag(sync=True)
    _synced_repr_model_ids = List().tag(sync=True)
    _ngl_view_id = List().tag(sync=True)
    _ngl_repr_dict = Dict().tag(sync=True)
    _ngl_component_ids = List().tag(sync=False)
    _ngl_component_names = List().tag(sync=False)
    _ngl_msg = None
    _send_binary = Bool(True).tag(sync=False)
    _init_gui = Bool(False).tag(sync=False)
    gui_style = CaselessStrEnum(['ngl'], allow_none=True).tag(sync=True)
    _gui_theme = CaselessStrEnum(['dark', 'light'], allow_none=True).tag(sync=True)
    _widget_theme = None
    _ngl_serialize = Bool(False).tag(sync=True)
    _ngl_msg_archive = List().tag(sync=True)
    _ngl_coordinate_resource = Dict().tag(sync=True)
    _representations = List().tag(sync=False)
    _ngl_color_dict = Dict().tag(sync=True)
    _player_dict = Dict().tag(sync=True)

    # instance
    _iplayer = Instance(widgets.Box,
                        allow_none=True).tag(sync=True, **widget_serialization)
    _igui = Instance(widgets.Tab,
                     allow_none=True).tag(sync=True, **widget_serialization)
    _ibtn_fullscreen = Instance(widgets.Button,
            allow_none=True).tag(sync=True, **widget_serialization)

    def __init__(self,
                 structure=None,
                 representations=None,
                 parameters=None,
                 **kwargs):
        super().__init__(**kwargs)

        self._gui = None
        self._init_gui = kwargs.pop('gui', False)
        self._theme = kwargs.pop('theme', 'default')
        self._widget_image = Image()
        self._widget_image.width = 900.
        self._image_array = []
        # do not use _displayed_callbacks since there is another Widget._display_callbacks
        self._event = threading.Event()
        self._ngl_displayed_callbacks_before_loaded = []
        widget_utils._add_repr_method_shortcut(self, self)
        self.shape = Shape(view=self)
        self.stage = Stage(view=self)
        self.control = ViewerControl(view=self)
        self._handle_msg_thread = threading.Thread(
            target=self.on_msg, args=(self._handle_custom_msg, ))
        # # register to get data from JS side
        self._handle_msg_thread.daemon = True
        self._handle_msg_thread.start()
        self._remote_call_thread = RemoteCallThread(
            self,
            registered_funcs=['loadFile', 'replaceStructure', '_exportImage'])
        self._remote_call_thread.start()
        self._trajlist = []
        self._ngl_component_ids = []

        if representations:
            # Must be set here before calling
            # add_trajectory or add_struture
            # After finish adding new Structure/Trajectory,
            # initial representations will be set.
            kwargs['default_representation'] = False
        else:
            if 'default' in kwargs:
                kwargs['default_representation'] = kwargs['default']

        autoview = 'center' not in kwargs or ('center' in kwargs
                                              and kwargs.pop('center'))
        # NOTE: Using `pop` to avoid passing `center` to NGL.

        if parameters:
            self.parameters = parameters
        if isinstance(structure, Trajectory):
            name = py_utils.get_name(structure, **kwargs)
            self.add_trajectory(structure, name=name, **kwargs)
        elif isinstance(structure, (list, tuple)):
            trajectories = structure
            for trajectory in trajectories:
                name = py_utils.get_name(trajectory, **kwargs)
                self.add_trajectory(trajectory, name=name, **kwargs)
        else:
            if structure is not None:
                self.add_structure(structure, **kwargs)

        if representations:
            # If initial representations are provided,
            # we need to set defaultRepresentation to False
            self.representations = representations
            if autoview:
                self.center()

        self.player = TrajectoryPlayer(self)
        self._view_width = kwargs.get('width', '')
        self._view_height = kwargs.get('height', '')

        # Updating only self.layout.{width, height} don't handle
        # resizing NGL widget properly.
        self._sync_with_layout()
        # self.layout.width = 'auto'
        self._create_player()
        self._create_ibtn_fullscreen()

    def _create_ibtn_fullscreen(self):
        button = widgets.Button(icon='compress')
        button.layout.width = '34px'
        # onclick is implemented in frontend
        self._ibtn_fullscreen = button


    def _sync_with_layout(self):
        def on_change_layout(change):
            new = change['new']
            if change['name'] == 'width':
                self._set_size(new, '')
            elif change['name'] == 'height':
                self._set_size('', new)

        self.layout.observe(on_change_layout, ['width', 'height'])

    def _set_serialization(self, frame_range=None):
        self._ngl_serialize = True
        resource = self._ngl_coordinate_resource
        if frame_range is not None:
            for t_index, traj in enumerate(self._trajlist):
                resource[t_index] = []
                for f_index in range(*frame_range):
                    if f_index < traj.n_frames:
                        resource[t_index].append(
                            encode_base64(traj.get_coordinates(f_index)))
                    else:
                        resource[t_index].append(
                            encode_base64(np.empty((0), dtype='f4')))
            resource['n_frames'] = len(resource[0])

        self._ngl_coordinate_resource = resource
        self._ngl_color_dict = color._USER_COLOR_DICT.copy()

    def _create_player(self):
        player = Play(max=self.max_frame, interval=100)
        slider = IntSlider(max=self.max_frame)
        self._iplayer = HBox([player, slider])
        self.player.widget_player = player
        self.player.widget_player_slider = slider

        jslink((player, 'value'), (slider, 'value'))
        jslink((player, 'value'), (self, 'frame'))
        jslink((player, 'max'), (self, 'max_frame'))
        jslink((slider, 'max'), (self, 'max_frame'))

    def _unset_serialization(self):
        self._ngl_serialize = False
        self._ngl_coordinate_resource = {}

    @property
    def parameters(self):
        return self._parameters

    @parameters.setter
    def parameters(self, params):
        params = _camelize_dict(params)
        self._parameters = params
        self._remote_call('setParameters', target='Widget', args=[
            params,
        ])

    @property
    def camera(self):
        return self._camera_str

    @camera.setter
    def camera(self, value):
        """
        
        Parameters
        ----------
        value : str, {'perspective', 'orthographic'}
        """
        self._camera_str = value
        # use _remote_call so this function can be called right after
        # self is displayed
        self._remote_call("setParameters",
                          target='Stage',
                          kwargs=dict(cameraType=self._camera_str))

    def _set_camera_orientation(self, arr):
        self._remote_call('set_camera_orientation',
                          target='Widget',
                          args=[
                              arr,
                          ])

    def _request_stage_parameters(self):
        self._remote_call('requestUpdateStageParameters', target='Widget')

    @validate('gui_style')
    def _validate_gui_style(self, proposal):
        val = proposal['value']
        if val == 'ngl':
            if self._widget_theme is None:
                from .theme import ThemeManager
                self._widget_theme = ThemeManager()
                _INIT_VIEWS['theme_mananager'] = self._widget_theme
                if self._widget_theme._theme is None:
                    self._widget_theme.light()
        return val

    @observe("_gui_theme")
    def _on_theme_changed(self, change):
        # EXPERIMENTAL
        from nglview.theme import theme
        if change.new == 'dark':
            self._widget_theme.dark()
        elif change.new == 'light':
            self._widget_theme.light()

    @observe('picked')
    def _on_picked(self, change):
        picked = change['new']
        if self.player.widget_picked is not None:
            self.player.widget_picked.value = json.dumps(picked)

    @observe('background')
    def _update_background_color(self, change):
        color = change['new']
        self.stage.set_parameters(background_color=color)

    def handle_resize(self):
        # self._remote_call("handleResize", target='Stage')
        self._remote_call("handleResize")

    @observe('n_components')
    def _handle_n_components_changed(self, change):
        if self.player.widget_repr is not None:
            component_slider = widget_utils.get_widget_by_name(
                self.player.widget_repr, 'component_slider')

            if change['new'] - 1 >= component_slider.min:
                component_slider.max = change['new'] - 1

            component_dropdown = widget_utils.get_widget_by_name(
                self.player.widget_repr, 'component_dropdown')
            component_dropdown.options = tuple(self._ngl_component_names)

            if change['new'] == 0:
                component_dropdown.options = tuple([' '])
                component_dropdown.value = ' '

                component_slider.max = 0

                reprlist_choices = widget_utils.get_widget_by_name(
                    self.player.widget_repr, 'reprlist_choices')
                reprlist_choices.options = tuple([' '])

                repr_slider = widget_utils.get_widget_by_name(
                    self.player.widget_repr, 'repr_slider')
                repr_slider.max = 0

                repr_name_text = widget_utils.get_widget_by_name(
                    self.player.widget_repr, 'repr_name_text')
                repr_selection = widget_utils.get_widget_by_name(
                    self.player.widget_repr, 'repr_selection')
                repr_name_text.value = ' '
                repr_selection.value = ' '

    @observe('_ngl_repr_dict')
    def _handle_repr_dict_changed(self, change):
        if self.player.widget_repr is not None:
            repr_slider = widget_utils.get_widget_by_name(
                self.player.widget_repr, 'repr_slider')
            component_slider = widget_utils.get_widget_by_name(
                self.player.widget_repr, 'component_slider')
            repr_name_text = widget_utils.get_widget_by_name(
                self.player.widget_repr, 'repr_name_text')
            repr_selection = widget_utils.get_widget_by_name(
                self.player.widget_repr, 'repr_selection')
            reprlist_choices = widget_utils.get_widget_by_name(
                self.player.widget_repr, 'reprlist_choices')
            repr_names = get_repr_names_from_dict(self._ngl_repr_dict,
                                                  component_slider.value)

            if change['new'] == {0: {}}:
                repr_selection.value = ''
            else:
                options = tuple(
                    str(i) + '-' + name for (i, name) in enumerate(repr_names))
                reprlist_choices.options = options

                try:
                    value = reprlist_choices.options[repr_slider.value]
                    if isinstance(value, tuple):
                        # https://github.com/jupyter-widgets/ipywidgets/issues/1512
                        value = value[0]
                    reprlist_choices.value = value
                except IndexError:
                    if repr_slider.value == 0:
                        # works fine with ipywidgets 5.2.2
                        reprlist_choices.options = tuple([
                            ' ',
                        ])
                        reprlist_choices.value = ' '
                    else:
                        reprlist_choices.value = reprlist_choices.options[
                            repr_slider.value - 1]

                # e.g: 0-cartoon
                repr_name_text.value = reprlist_choices.value.split(
                    '-')[-1].strip()

                repr_slider.max = len(repr_names) - 1 if len(
                    repr_names) >= 1 else len(repr_names)

    def _update_max_frame(self):
        self.max_frame = max(
            int(traj.n_frames) for traj in self._trajlist
            if hasattr(traj, 'n_frames')) - 1 # index starts from 0

    def _wait_until_finished(self, timeout=0.0001):
        # NGL need to send 'finished' signal to
        # backend
        self._event.clear()
        while True:
            # idle to make room for waiting for
            # "finished" event sent from JS
            time.sleep(timeout)
            if self._event.is_set():
                # if event is set from another thread
                # break while True
                break

    def _run_on_another_thread(self, func, *args):
        # use `event` to singal
        # func(*args)
        thread = threading.Thread(
            target=func,
            args=args,
        )
        thread.daemon = True
        thread.start()
        return thread

    @observe('loaded')
    def on_loaded(self, change):
        # trick for firefox on Linux
        time.sleep(0.1)

        if change['new']:
            self._fire_callbacks(self._ngl_displayed_callbacks_before_loaded)

    def _fire_callbacks(self, callbacks):
        def _call(event):
            for callback in callbacks:
                callback(self)
                if callback._method_name == 'loadFile':
                    self._wait_until_finished()

        self._run_on_another_thread(_call, self._event)

    def _ipython_display_(self, **kwargs):
        try:
            # ipywidgets < 8
            super()._ipython_display_(**kwargs)
        except AttributeError:
            display(super()._repr_mimebundle_(), raw=True)
        if self._init_gui:
            if self._gui is None:
                self._gui = self.player._display()
            display(self._gui)

    def display(self, gui=False, style='ngl'):
        """

        Parameters
        ----------
        gui : bool
            If True: turn on GUI
        style : str, {'ngl', 'ipywidgets}, default 'ngl'
            GUI style (with gui=True)
        """
        if gui:
            if style == 'ipywidgets':
                # For the old implementation
                # is there anyone using this?
                self.gui_style = None # turn off the NGL's GUI
                self._gui = self.player._display()
                self._gui.layout.align_self = 'stretch'
                self._gui.layout.width = '400px'
                b = HBox([self, self._gui])
                def on(b):
                    self.handle_resize()
                try:
                    # ipywidgets < 8
                    b.on_displayed(on)
                except AttributeError:
                    logger.warn(
                        "display(style='ipywidgets') is not supported"
                        " with this version of ipywidgets"
                    )
                return b
            elif style == 'ngl':
                self.gui_style = 'ngl'
                return self
        else:
            return self

    def _set_size(self, w, h):
        '''

        Parameters
        ----------
        w, h : float or str

        Examples
        --------
        >>> import nglview; view = nglview.demo()
        >>> view._set_size(100, 100)
        >>> view._set_size('100px', '100px')
        >>> view._set_size('50%', '50%')
        '''
        self._remote_call('setSize', target='Widget', args=[w, h])

    def _set_sync_repr(self, other_views):
        model_ids = {v._model_id for v in other_views}
        self._synced_repr_model_ids = sorted(
            set(self._synced_repr_model_ids) | model_ids)
        self._remote_call("setSyncRepr",
                          target="Widget",
                          args=[self._synced_repr_model_ids])

    def _set_unsync_repr(self, other_views):
        model_ids = {v._model_id for v in other_views}
        self._synced_repr_model_ids = list(set(self._synced_repr_model_ids) - model_ids)
        self._remote_call("setSyncRepr",
                          target="Widget",
                          args=[self._synced_repr_model_ids])

    def _set_sync_camera(self, other_views):
        model_ids = {v._model_id for v in other_views}
        self._synced_model_ids = sorted(
            set(self._synced_model_ids) | model_ids)
        self._remote_call("setSyncCamera",
                          target="Widget",
                          args=[self._synced_model_ids])

    def _set_unsync_camera(self, other_views):
        model_ids = {v._model_id for v in other_views}
        self._synced_model_ids = list(set(self._synced_model_ids) - model_ids)
        self._remote_call("setSyncCamera",
                          target="Widget",
                          args=[self._synced_model_ids])

    def _set_spin(self, axis, angle):
        self._remote_call('setSpin', target='Stage', args=[axis, angle])

    def _set_selection(self, selection, component=0, repr_index=0):
        self._remote_call("setSelection",
                          target='Representation',
                          args=[selection],
                          kwargs=dict(component_index=component,
                                      repr_index=repr_index))

    def color_by(self, color_scheme, component=0):
        '''update color for all representations of given component

        Notes
        -----
        Unstable feature

        Parameters
        ----------
        color_scheme : str
        component : int, default 0
            component index

        Examples
        --------
        >>> import nglview
        >>> view = nglview.demo()
        >>> # component 0
        >>> view.color_by('atomindex')

        >>> # component 1
        >>> view.color_by('atomindex', component=1)
        '''
        repr_names = get_repr_names_from_dict(self._ngl_repr_dict, component)

        for index, _ in enumerate(repr_names):
            self.update_representation(component=component,
                                       repr_index=index,
                                       color_scheme=color_scheme)

    @property
    def representations(self):
        return self._representations

    @representations.setter
    def representations(self, reps):
        if isinstance(reps, dict):
            self._remote_call("_set_representation_from_repr_dict",
                    args=[reps])
        else:
            self._representations = reps[:]
            for index in range(len(self._ngl_component_ids)):
                self.set_representations(reps)

    def update_representation(self, component=0, repr_index=0, **parameters):
        """

        Parameters
        ----------
        component : int, default 0
            component index
        repr_index : int, default 0
            representation index for given component
        parameters : dict
        """
        parameters = _camelize_dict(parameters)
        kwargs = dict(component_index=component, repr_index=repr_index)
        kwargs.update(parameters)

        self._remote_call('setParameters',
                          target='Representation',
                          kwargs=kwargs)
        self._update_repr_dict()

    def _update_repr_dict(self):
        """ Send a request to fronend to send representation parameters
        back.

        # TODO: sync or async
        """
        self._remote_call('request_repr_dict', target='Widget')

    def set_representations(self, representations, component=0):
        """
        
        Parameters
        ----------
        representations : list of dict
        """
        self.clear_representations(component=component)

        for params in representations:
            assert isinstance(params, dict), 'params must be a dict'
            kwargs = params['params']
            kwargs.update({'component_index': component})
            self._remote_call('addRepresentation',
                              target='compList',
                              args=[
                                  params['type'],
                              ],
                              kwargs=kwargs)

    def _remove_representation(self, component=0, repr_index=0):
        self._remote_call('removeRepresentation',
                          target='Widget',
                          args=[component, repr_index])

    def _remove_representations_by_name(self, repr_name, component=0):
        self._remote_call('removeRepresentationsByName',
                          target='Widget',
                          args=[repr_name, component])

    def _update_representations_by_name(self, repr_name, component=0,
                                        **kwargs):
        kwargs = _camelize_dict(kwargs)
        self._remote_call('updateRepresentationsByName',
                          target='Widget',
                          args=[repr_name, component],
                          kwargs=kwargs)

    def _display_repr(self, component=0, repr_index=0, name=None):
        c = 'c' + str(component)
        r = str(repr_index)

        try:
            name = self._ngl_repr_dict[c][r]['type']
        except KeyError:
            name = ''

        return RepresentationControl(self, component, repr_index, name=name)

    def _set_coordinates(self, index, movie_making=False, render_params=None):
        # FIXME: use movie_making here seems awkward.
        '''update coordinates for all trajectories at index-th frame
        '''
        render_params = render_params or {}
        if self._trajlist:
            coordinates_dict = {}
            for trajectory in self._trajlist:
                traj_index = self._ngl_component_ids.index(trajectory.id)

                try:
                    if trajectory.shown:
                        if self.player.interpolate:
                            t = self.player.iparams.get('t', 0.5)
                            step = self.player.iparams.get('step', 1)
                            coordinates_dict[traj_index] = interpolate.linear(
                                index, t=t, traj=trajectory, step=step)
                        else:
                            coordinates_dict[
                                traj_index] = trajectory.get_coordinates(index)
                    else:
                        coordinates_dict[traj_index] = np.empty((0),
                                                                dtype='f4')
                except (IndexError, ValueError):
                    coordinates_dict[traj_index] = np.empty((0), dtype='f4')

            self.set_coordinates(coordinates_dict,
                    render_params=render_params,
                    movie_making=movie_making)
        else:
            print("no trajectory available")

    def set_coordinates(self, arr_dict, movie_making=False,
            render_params=None):
        # type: (Dict[int, np.ndarray]) -> None
        """Used for update coordinates of a given trajectory
        >>> # arr: numpy array, ndim=2
        >>> # update coordinates of 1st trajectory
        >>> view.set_coordinates({0: arr})# doctest: +SKIP
        """
        render_params = render_params or {}
        self._coordinates_dict = arr_dict

        buffers = []
        coordinates_meta = dict()
        for index, arr in self._coordinates_dict.items():
            buffers.append(arr.astype('f4').tobytes())
            coordinates_meta[index] = index
        msg = {
                'type': 'binary_single',
                'data': coordinates_meta,
            }
        if movie_making:
            msg['movie_making'] = movie_making
            msg['render_params'] = render_params

        self.send(
            msg,
            buffers=buffers)

    @observe('frame')
    def _on_frame_changed(self, change):
        """set and send coordinates at current frame
        """
        self._set_coordinates(self.frame)

    def clear(self, *args, **kwargs):
        '''shortcut of `clear_representations`
        '''

        self.clear_representations(*args, **kwargs)

    def clear_representations(self, component=0):
        '''clear all representations for given component

        Parameters
        ----------
        component : int, default 0 (first model)
            You need to keep track how many components you added.
        '''
        self._remote_call("removeAllRepresentations",
                          target='compList',
                          kwargs={'component_index': component})

    @_update_url
    def _add_shape(self, shapes, name='shape'):
        """add shape objects

        TODO: update doc, caseless shape keyword

        Parameters
        ----------
        shapes : list of tuple
        name : str, default 'shape'
            name of given shape

        Notes
        -----
        Supported shape: 'mesh', 'sphere', 'ellipsoid', 'cylinder', 'cone', 'arrow'.
        
        See also
        --------
        {ngl_url}

        Examples
        --------
        >>> import nglview
        >>> view = nglview.demo()
        >>> sphere = ('sphere', [0, 0, 9], [1, 0, 0], 1.5)
        >>> arrow = ('arrow', [1, 2, 7 ], [30, 3, 3], [1, 0, 1], 1.0)
        >>> c = view._add_shape([sphere, arrow], name='my_shape')
        """

        self._remote_call('addShape', target='Widget', args=[name, shapes], fire_embed=True)

        # Added to remain in sync with the JS components
        # Similarly to _loadData
        cid = str(uuid.uuid4())
        self._ngl_component_ids.append(cid)

        comp_name = py_utils.get_name(self.shape)
        self._ngl_component_names.append(comp_name)

        self._update_component_auto_completion()
        return ComponentViewer(self, cid)

    @_update_url
    def add_representation(self, repr_type, selection='all', **kwargs):
        '''Add structure representation (cartoon, licorice, ...) for given atom selection.

        Parameters
        ----------
        repr_type : str
            type of representation. Please see {ngl_url} for further info.
        selection : str or 1D array (atom indices) or any iterator that returns integer, default 'all'
            atom selection
        **kwargs: additional arguments for representation

        Example
        -------
        >>> import nglview as nv
        >>> w = nv.demo()
        >>> w.add_representation('cartoon', selection='protein', color='blue')
        >>> w.add_representation('licorice', selection=[3, 8, 9, 11], color='red')
        >>> w # doctest: +SKIP

        Notes
        -----
        User can also use shortcut

        >>> selection = 'protein'
        >>> w.add_cartoon(selection) # w.add_representation('cartoon', selection)
        '''
        if repr_type == 'surface':
            if 'useWorker' not in kwargs:
                kwargs['useWorker'] = False

        # avoid space sensitivity
        repr_type = repr_type.strip()
        # overwrite selection
        selection = seq_to_string(selection).strip()

        # make copy
        kwargs2 = _camelize_dict(kwargs)

        if 'component' in kwargs2:
            component = kwargs2.pop('component')
        else:
            component = 0

        for k, v in kwargs2.items():
            try:
                kwargs2[k] = v.strip()
            except AttributeError:
                # e.g.: opacity=0.4
                kwargs2[k] = v

        d = {'params': {'sele': selection}}
        d['type'] = repr_type
        d['params'].update(kwargs2)

        params = d['params']
        params.update({'component_index': component})
        self._remote_call('addRepresentation',
                          target='compList',
                          args=[
                              d['type'],
                          ],
                          kwargs=params)

    @_deprecated("DEPRECATED: Please use 'center' method")
    def center_view(self, *args, **kwargs):
        """alias of `center_view`
        """
        self.center(*args, **kwargs)

    def center(self, selection='*', duration=0, component=0, **kwargs):
        """center view for given atom selection

        Examples
        --------
        view.center_view(selection='1-4')
        """
        self._remote_call('autoView',
                          target='compList',
                          args=[selection, duration],
                          kwargs={'component_index': component},
                          **kwargs)

    @observe('_image_data')
    def _on_render_image(self, change):
        '''update image data to widget_image

        Notes
        -----
        method name might be changed
        '''
        self._widget_image._b64value = change['new']

    def render_image(self,
                     frame=None,
                     factor=4,
                     antialias=True,
                     trim=False,
                     transparent=False):
        """render and get image as ipywidgets.widget_image.Image

        Parameters
        ----------
        frame : int or None, default None
            if None, use current frame
            if specified, use this number.
        factor : int, default 4
            quality of the image, higher is better
        antialias : bool, default True
        trim : bool, default False
        transparent : bool, default False

        Examples
        --------
            # tell NGL to render send image data to notebook.
            view.render_image()
            
            # make sure to call `get_image` method
            view.get_image()

        Notes
        -----
        You need to call `render_image` and `get_image` in different notebook's Cells
        """
        if frame is not None:
            self.frame = frame
        params = dict(factor=factor,
                      antialias=antialias,
                      trim=trim,
                      transparent=transparent)
        iw = Image()
        iw.width = '99%'  # avoid ugly scroll bar on notebook.
        self._remote_call('_exportImage',
                          target='Widget',
                          args=[iw.model_id],
                          kwargs=params)
        # iw.value will be updated later after frontend send the image_data back.
        _TRACKED_WIDGETS[iw.model_id] = iw
        return iw

    def download_image(self,
                       filename='screenshot.png',
                       factor=4,
                       antialias=True,
                       trim=False,
                       transparent=False):
        """render and download scene at current frame

        Parameters
        ----------
        filename : str, default 'screenshot.png'
        factor : int, default 4
            quality of the image, higher is better
        antialias : bool, default True
        trim : bool, default False
        transparent : bool, default False
        """
        params = dict(factor=factor,
                      antialias=antialias,
                      trim=trim,
                      transparent=transparent)
        self._remote_call('_downloadImage',
                          target='Widget',
                          args=[
                              filename,
                          ],
                          kwargs=params)

    def _handle_custom_msg(self, msg, buffers):
        """store message sent from Javascript.

        How? use view.on_msg(get_msg)

        Notes: message format should be {'type': type, 'data': data}
        _handle_custom_msg will call appropriate function to handle message "type"
        """
        self._ngl_msg = msg

        msg_type = self._ngl_msg.get('type')
        if msg_type == 'request_frame':
            frame = self.frame + self.player.step
            if frame > self.max_frame:
                frame = 0
            elif frame < 0:
                frame = self.max_frame
            self.frame = frame
        elif msg_type == 'updateIDs':
            self._ngl_view_id = msg['data']
        elif msg_type == 'removeComponent':
            cindex = int(msg['data'])
            self._ngl_component_ids.pop(cindex)
        elif msg_type == 'repr_parameters':
            data_dict = self._ngl_msg.get('data')
            name = data_dict.pop('name') + '\n'
            selection = data_dict.get('sele', '') + '\n'
            # json change True to true
            data_dict_json = json.dumps(data_dict).replace(
                'true', 'True').replace('false', 'False')
            data_dict_json = data_dict_json.replace('null', '"null"')

            if self.player.widget_repr is not None:
                # TODO: refactor
                repr_name_text = widget_utils.get_widget_by_name(
                    self.player.widget_repr, 'repr_name_text')
                repr_selection = widget_utils.get_widget_by_name(
                    self.player.widget_repr, 'repr_selection')
                repr_name_text.value = name
                repr_selection.value = selection
        elif msg_type == 'request_loaded':
            if not self.loaded:
                # trick to trigger observe loaded
                # so two viewers can have the same representations
                self.loaded = False
            self.loaded = msg.get('data')
        elif msg_type == 'request_repr_dict':
            # update _repr_dict will trigger other things
            # see _handle_repr_dict_changed
            self._ngl_repr_dict = self._ngl_msg.get('data')
        elif msg_type == 'stage_parameters':
            self._ngl_full_stage_parameters = msg.get('data')
        elif msg_type == 'async_message':
            if msg.get('data') == 'ok':
                self._event.set()
        elif msg_type == 'image_data':
            self._image_data = msg.get('data')
            _TRACKED_WIDGETS[msg.get('ID')].value = base64.b64decode(
                self._image_data)

    def _request_repr_parameters(self, component=0, repr_index=0):
        if self.n_components > 0:
            self._remote_call('requestReprParameters',
                              target='Widget',
                              args=[component, repr_index])

    def add_structure(self, structure, **kwargs):
        '''add structure to view

        Parameters
        ----------
        structure : nglview.Structure object

        Examples
        --------
        >>> view.add_trajectory(traj0) # doctest: +SKIP
        ... view.add_trajectory(traj1)
        ... # then add Structure
        ... view.add_structure(s)

        See Also
        --------
        nglview.NGLWidget.add_component
        '''
        if not isinstance(structure, Structure):
            raise ValueError(f'{structure} is not an instance of Structure')
        self._load_data(structure, **kwargs)
        self._ngl_component_ids.append(structure.id)
        if self.n_components > 1:
            self.center_view(component=len(self._ngl_component_ids) - 1)
        self._update_component_auto_completion()
        return self[-1]

    def add_trajectory(self, trajectory, **kwargs):
        '''add new trajectory to `view`

        Parameters
        ----------
        trajectory: nglview.Trajectory or its derived class or
            a supported object, eg pytraj.Trajectory-like,
            mdtraj.Trajectory, MDAnalysis objects, etc

        See Also
        --------
        nglview.NGLWidget.add_component

        Examples
        --------
        >>> import nglview as nv
        >>> traj = nv.SimpletrajTrajectory(nv.datafiles.XTC, nv.datafiles.PDB)
        >>> view = nv.NGLWidget()
        >>> c = view.add_trajectory(traj)
        '''
        backends = BACKENDS

        package_name = trajectory.__module__.split('.')[0]

        if package_name in backends:
            trajectory = backends[package_name](trajectory)
        else:
            trajectory = trajectory

        self._load_data(trajectory, **kwargs)
        setattr(trajectory, 'shown', True)
        self._trajlist.append(trajectory)
        self._update_max_frame()
        self._ngl_component_ids.append(trajectory.id)
        self._update_component_auto_completion()
        return self[-1]

    def add_pdbid(self, pdbid, **kwargs):
        '''add new Structure view by fetching pdb id from rcsb

        Examples
        --------
        >>> import nglview
        >>> view = nglview.NGLWidget()
        >>> c = view.add_pdbid('1tsu')
        >>> # which is equal to 
        >>> # view.add_component('rcsb://1tsu.pdb')
        '''
        return self.add_component(f'rcsb://{pdbid}.pdb', **kwargs)

    def add_component(self, filename, **kwargs):
        '''add component from file/trajectory/struture

        Parameters
        ----------
        filename : str or Trajectory or Structure or their derived class or url
        **kwargs : additional arguments, optional

        Examples
        --------
        >>> import nglview
        >>> view = nglview.NGLWidget()
        >>> view # doctest: +SKIP
        ... filename = 'somefile.ccp4'
        ... view.add_component(filename)

        Notes
        -----
        If you want to load binary file such as density data, mmtf format, it is
        faster to load file from current or subfolder.
        '''
        # if passed a supported object, convert "filename" to nglview.Trajectory
        try:
            package_name = filename.__module__.split('.')[0]
        except (TypeError, AttributeError):
            # string filename
            pass
        else:
            if package_name in BACKENDS:
                filename = BACKENDS[package_name](filename)

        self._load_data(filename, **kwargs)
        # assign an ID
        self._ngl_component_ids.append(str(uuid.uuid4()))
        self._update_component_auto_completion()
        return self[-1]

    def _load_data(self, obj, **kwargs):
        '''

        Parameters
        ----------
        obj : nglview.Structure or any object having 'get_structure_string' method or
              string buffer (open(fn).read())
        '''
        kwargs2 = _camelize_dict(kwargs)

        try:
            is_url = FileManager(obj).is_url
        except NameError:
            is_url = False

        if 'defaultRepresentation' not in kwargs2:
            kwargs2['defaultRepresentation'] = True

        if not is_url:
            if hasattr(obj, 'get_structure_string'):
                blob = obj.get_structure_string()
                kwargs2['ext'] = obj.ext
                passing_buffer = True
                binary = False
            else:
                fh = FileManager(obj,
                                 ext=kwargs.get('ext'),
                                 compressed=kwargs.get('compressed'))
                # assume passing string
                blob = fh.read()
                passing_buffer = not fh.use_filename

                if fh.ext is None and passing_buffer:
                    raise ValueError('must provide extension')

                kwargs2['ext'] = fh.ext
                binary = fh.is_binary
                use_filename = fh.use_filename

            if binary and not use_filename:
                # send base64
                blob = base64.b64encode(blob).decode('utf8')
            blob_type = 'blob' if passing_buffer else 'path'
            args = [{'type': blob_type, 'data': blob, 'binary': binary}]
        else:
            # is_url
            blob_type = 'url'
            url = obj
            args = [{'type': blob_type, 'data': url, 'binary': False}]

        name = py_utils.get_name(obj, **kwargs2)
        self._ngl_component_names.append(name)
        self._remote_call("loadFile",
                          target='Stage',
                          args=args,
                          kwargs=kwargs2)

    def remove_component(self, c):
        """remove component by its uuid.
        If isinstance(c, ComponentViewer), `c` won't be associated with `self`

        Parameters
        ----------
        c : Union[int, ComponentViewer]

        Examples
        --------
        >>> c0 = view.add_trajectory(traj0) # doctest: +SKIP
        ... c1 = view.add_trajectory(traj1)
        ... c2 = view.add_struture(structure)
        ... # remove last component
        ... view.remove_component(c2)
        ... assert c2._view is None
        """
        if isinstance(c, ComponentViewer):
            component_id = c.id
            c._view = None
        else:
            component_id = c
        self._clear_component_auto_completion()
        if self._trajlist:
            for traj in self._trajlist:
                if traj.id == component_id:
                    self._trajlist.remove(traj)
        component_index = self._ngl_component_ids.index(component_id)
        self._ngl_component_ids.remove(component_id)
        self._ngl_component_names.pop(component_index)

        self._remote_call('removeComponent',
                          target='Stage',
                          args=[
                              component_index,
                          ])

        self._update_component_auto_completion()

    def _dry_run(self, func, *args, **kwargs):
        return _dry_run(self, func, *args, **kwargs)

    def _get_remote_call_msg(self,
                             method_name,
                             target='Widget',
                             args=None,
                             kwargs=None,
                             **other_kwargs):
        """call NGL's methods from Python.
        
        Parameters
        ----------
        method_name : str
        target : str, {'Stage', 'Viewer', 'compList', 'StructureComponent'}
        args : list
        kwargs : dict
            if target is 'compList', "component_index" could be passed
            to specify which component will call the method.

        Examples
        --------
        view._remote_call('loadFile', args=['1L2Y.pdb'],
                          target='Stage', kwargs={'defaultRepresentation': True})

        # perform autoView for 1st component
        # JS code
        # component = Stage.compList[1];
        # component.autoView('*', 200)

        # python
        view._remote_call('autoView',
                          target='component',
                          args=['*', 200],
                          kwargs={'component_index': 1})
        """
        # NOTE: _camelize_dict here?
        args = [] if args is None else args
        kwargs = {} if kwargs is None else kwargs

        msg = {}

        if 'component_index' in kwargs:
            msg['component_index'] = kwargs.pop('component_index')
        if 'repr_index' in kwargs:
            msg['repr_index'] = kwargs.pop('repr_index')
        if 'default' in kwargs:
            kwargs['defaultRepresentation'] = kwargs.pop('default')

        # Color handling
        reconstruc_color_scheme = False
        if 'color' in kwargs and isinstance(kwargs['color'],
                                            color._ColorScheme):
            kwargs['color_label'] = kwargs['color'].data['label']
            # overite `color`
            kwargs['color'] = kwargs['color'].data['data']
            reconstruc_color_scheme = True
        if kwargs.get('colorScheme') == 'volume' and kwargs.get('colorVolume'):
            assert isinstance(kwargs['colorVolume'], ComponentViewer)
            kwargs['colorVolume'] = kwargs['colorVolume']._index

        msg['target'] = target
        msg['type'] = 'call_method'
        msg['methodName'] = method_name
        msg['reconstruc_color_scheme'] = reconstruc_color_scheme
        msg['args'] = args
        msg['kwargs'] = kwargs
        if other_kwargs:
            msg.update(other_kwargs)
        return msg

    def _trim_message(self, messages):
        messages = messages[:]

        remove_comps = [(index, msg['args'][0])
                        for index, msg in enumerate(messages)
                        if msg['methodName'] == 'removeComponent']

        if not remove_comps:
            return messages

        load_comps = [
            index for index, msg in enumerate(messages)
            if msg['methodName'] in ('loadFile', 'addShape')
        ]

        messages_rm = [r[0] for r in remove_comps]
        messages_rm += [load_comps[r[1]] for r in remove_comps]
        messages_rm = set(messages_rm)

        return [
            msg for i, msg in enumerate(messages)
            if i not in messages_rm
        ]

    def _remote_call(self,
                     method_name,
                     target='Widget',
                     args=None,
                     kwargs=None,
                     **other_kwargs):

        msg = self._get_remote_call_msg(method_name,
                                        target=target,
                                        args=args,
                                        kwargs=kwargs,
                                        **other_kwargs)

        def callback(widget, msg=msg):
            widget.send(msg)

        callback._method_name = method_name
        callback._ngl_msg = msg

        if self.loaded:
            self._remote_call_thread.q.append(callback)
        else:
            # send later
            # all callbacks will be called right after widget is loaded
            self._ngl_displayed_callbacks_before_loaded.append(callback)

        if callback._method_name not in _EXCLUDED_CALLBACK_AFTER_FIRING and \
           (not other_kwargs.get("fire_once", False)):
            archive = self._ngl_msg_archive[:]
            archive.append(msg)
            self._ngl_msg_archive = self._trim_message(archive)

    def _get_traj_by_id(self, itsid):
        """return nglview.Trajectory or its derived class object
        """
        for traj in self._trajlist:
            if traj.id == itsid:
                return traj
        return None

    def hide(self, indices):
        """set invisibility for given component/struture/trajectory (by their indices)
        """
        traj_ids = {traj.id for traj in self._trajlist}

        for index in indices:
            comp_id = self._ngl_component_ids[index]
            if comp_id in traj_ids:
                traj = self._get_traj_by_id(comp_id)
                traj.shown = False
            self._remote_call("setVisibility",
                              target='compList',
                              args=[
                                  False,
                              ],
                              kwargs={'component_index': index})

    def show(self, **kwargs):
        """shortcut of `show_only`
        """
        self.show_only(**kwargs)

    def show_only(self, indices='all', **kwargs):
        """set visibility for given components (by their indices)

        Parameters
        ----------
        indices : {'all', array-like}, component index, default 'all'
        """
        traj_ids = {traj.id for traj in self._trajlist}

        if indices == 'all':
            indices_ = set(range(self.n_components))
        else:
            indices_ = set(indices)

        for index, comp_id in enumerate(self._ngl_component_ids):
            if comp_id in traj_ids:
                traj = self._get_traj_by_id(comp_id)
            else:
                traj = None
            if index in indices_:
                args = [
                    True,
                ]
                if traj is not None:
                    traj.shown = True
            else:
                args = [
                    False,
                ]
                if traj is not None:
                    traj.shown = False

            self._remote_call("setVisibility",
                              target='compList',
                              args=args,
                              kwargs={'component_index': index},
                              **kwargs)

    def _js_console(self):
        self.send(dict(type='get', data='any'))

    def _get_full_params(self):
        self.send(dict(type='get', data='parameters'))

    def _display_image(self):
        '''for testing
        '''
        from IPython import display
        im_bytes = base64.b64decode(self._image_data)
        return display.Image(im_bytes)

    def _clear_component_auto_completion(self):
        for index, _ in enumerate(self._ngl_component_ids):
            name = 'component_' + str(index)
            delattr(self, name)

    def _js(self, code, **kwargs):
        self._execute_js_code(code, **kwargs)

    def _execute_js_code(self, code, **kwargs):
        self._remote_call('executeCode',
                          target='Widget',
                          args=[code],
                          **kwargs)

    def _update_component_auto_completion(self):
        trajids = [traj.id for traj in self._trajlist]

        for index, cid in enumerate(self._ngl_component_ids):
            comp = ComponentViewer(self, cid)
            name = 'component_' + str(index)
            setattr(self, name, comp)

            if cid in trajids:
                traj_name = 'trajectory_' + str(trajids.index(cid))
                setattr(self, traj_name, comp)

    def __getitem__(self, index):
        """return ComponentViewer
        """
        postive_index = py_utils.get_positive_index(
            index, len(self._ngl_component_ids))
        return ComponentViewer(self, self._ngl_component_ids[postive_index])

    def __iter__(self):
        """return ComponentViewer
        """
        for i, _ in enumerate(self._ngl_component_ids):
            yield self[i]


class Fullscreen(DOMWidget):
    """EXPERIMENTAL
    """
    _view_name = Unicode("FullscreenView").tag(sync=True)
    _view_module = Unicode("nglview-js-widgets").tag(sync=True)
    _view_module_version = Unicode(__frontend_version__).tag(sync=True)
    _model_name = Unicode("FullscreenModel").tag(sync=True)
    _model_module = Unicode("nglview-js-widgets").tag(sync=True)
    _model_module_version = Unicode(__frontend_version__).tag(sync=True)

    _is_fullscreen = Bool().tag(sync=True)

    def __init__(self, target, views):
        super().__init__()
        self._target = target
        self._views = views

    def fullscreen(self):
        self._js("this.fullscreen('%s')" % self._target.model_id)

    def _js(self, code):
        msg = {"executeCode": code}
        self.send(msg)

    @observe('_is_fullscreen')
    def _fullscreen_changed(self, change):
        if not change.new:
            self._target.layout.height = '300px'
        self.handle_resize()

    def handle_resize(self):
        for v in self._views:
            v.handle_resize()



import gzip
import os
import sys
import time
import unittest
from functools import partial
from io import StringIO
from itertools import chain

import ipywidgets
import numpy as np
import pytest
import traitlets
from ipykernel.comm import Comm
from IPython.display import display
from ipywidgets import BoundedFloatText, Button, HBox, IntText, Layout, Widget
from mock import MagicMock, patch
from numpy.testing import assert_almost_equal as aa_eq
from traitlets import TraitError, link

import nglview as nv
from nglview import NGLWidget, interpolate, js_utils, widget_utils
from nglview.representation import RepresentationControl
from nglview.utils.py_utils import click, decode_base64, encode_base64, submit
# local
from utils import get_fn
from utils import repr_dict as REPR_DICT

try:
    import simpletraj
    has_simpletraj = True
except ImportError:
    has_simpletraj = False

try:
    import pytraj as pt
    has_pytraj = True
except ImportError:
    pt = None
    has_pytraj = False

try:
    import mdtraj as md
    has_mdtraj = True
except ImportError:
    has_mdtraj = False
try:
    import parmed as pmd
    has_parmed = True
except ImportError:
    has_parmed = False

try:
    import MDAnalysis
    has_MDAnalysis = True
except ImportError:
    has_MDAnalysis = False

try:
    import htmd
    has_HTMD = True
except ImportError:
    has_HTMD = False

try:
    import ase
    has_ase = True
except ImportError:
    has_ase = False

try:
    import pymatgen
    has_pymatgen = True
except ImportError:
    has_pymatgen = False

try:
    import Bio.PDB
    has_bio = True
except ImportError:
    has_bio = False

try:
    import qcelemental
    has_qcelemental = True
except ImportError:
    has_qcelemental = False


def get_simple_traj():
    return nv.SimpletrajTrajectory(nv.datafiles.XTC, nv.datafiles.GRO)


def default_view():
    view = nv.NGLWidget()
    view.add_trajectory(get_simple_traj())
    return view


#-----------------------------------------------------------------------------
# NGLView stuff
#-----------------------------------------------------------------------------

DEFAULT_REPR = [{
    'params': {
        'sele': 'polymer'
    },
    'type': 'cartoon'
}, {
    'params': {
        'sele': 'hetero OR mol'
    },
    'type': 'ball+stick'
}, {
    "type": "ball+stick",
    "params": {
        "sele": "not protein and not nucleic"
    }
}]


def _assert_dict_list_equal(listdict0, listdict1):
    for (dict0, dict1) in zip(listdict0, listdict1):
        for (key0, key1) in zip(sorted(dict0.keys()), sorted(dict1.keys())):
            assert key0 == key1
            assert dict0.get(key0) == dict1.get(key1)


def test_API_promise_to_have():

    # for Jupyter notebook extension
    nv._jupyter_nbextension_paths()

    view = nv.demo()

    # trigger _set_size
    with patch.object(view, '_remote_call') as mock_call:
        view.layout.width = '100px'
        view.layout.height = '500px'
        mock_call.assert_called_with('setSize',
                                     args=['', '500px'],
                                     target='Widget')

    # Structure
    structure = nv.Structure()
    structure.get_structure_string
    assert hasattr(structure, 'id')
    assert hasattr(structure, 'ext')
    assert hasattr(structure, 'params')

    # Widget
    nv.NGLWidget._set_coordinates

    nv.NGLWidget.add_component
    nv.NGLWidget.add_trajectory
    nv.NGLWidget._coordinates_dict
    nv.NGLWidget.set_representations
    nv.NGLWidget.clear
    nv.NGLWidget.center

    # add component
    view.add_component('rcsb://1tsu.pdb')
    view.add_pdbid('1tsu')

    # display
    js_utils.clean_error_output()
    display(view.player.widget_repr)
    view.player._display()
    view._display_image()

    # show
    try:
        nv.show_pdbid('1tsu')
    except:
        pass
    nv.show_url('https://dummy.pdb')
    # other backends will be tested in other sections

    # constructor
    ngl_traj = get_simple_traj()
    nv.NGLWidget(ngl_traj, parameters=dict(background_color='black'))
    nv.NGLWidget(ngl_traj, representations=[dict(type='cartoon', params={})])

    view.parameters
    view.camera
    view.camera = 'perspective'
    view._request_stage_parameters()
    view._ngl_repr_dict = REPR_DICT
    view._handle_repr_dict_changed(dict(new=dict(c0={})))

    # dummy
    class DummWidget():
        value = ''

    view.player.picked_widget = DummWidget()

    view._update_background_color(change=dict(new='blue'))
    tab = view.player._display()

    view.player.widget_repr = view.player._make_widget_repr()
    view._handle_n_components_changed(change=dict(new=2, old=1))
    view._handle_n_components_changed(change=dict(new=1, old=1))
    view._handle_n_components_changed(change=dict(new=1, old=0))
    view.on_loaded(change=dict(new=True))
    view.on_loaded(change=dict(new=False))

    view._first_time_loaded = False
    view
    view._first_time_loaded = True
    view
    view._init_gui = True
    view
    view._theme = 'dark'
    view

    view.display(gui=True, style='ngl')
    view.display(gui=False)
    view.display(gui=True, style='ipywidgets')
    view._set_sync_camera([view])
    view._set_unsync_camera([view])
    view._set_selection('.CA')
    view.color_by('atomindex')
    representations = [dict(type='cartoon', params=dict())]
    view.representations = representations
    repr_parameters = dict(opacity=0.3, params=dict())
    view.update_representation(parameters=repr_parameters)
    view._remove_representation()
    view.clear()
    view.add_representation('surface', selection='*', useWorker=True)
    view.add_representation('surface', selection='*', component=1)
    view.center()
    view._on_render_image(change=dict(new='xyz'))
    view.render_image()
    view.render_image(frame=2)
    view.download_image()

    assert view._dry_run(view._set_sync_camera,
                         [view])['methodName'] == 'setSyncCamera'

    msg = dict(type='request_frame', data=dict())
    view._handle_custom_msg(msg=msg, buffers=[])
    msg = dict(type='repr_parameters', data=dict(name='hello'))
    view._handle_custom_msg(msg=msg, buffers=[])
    view.loaded = True
    msg = dict(type='request_loaded', data=True)
    view._handle_custom_msg(msg=msg, buffers=[])
    view.loaded = False
    msg = dict(type='request_loaded', data=True)
    view._handle_custom_msg(msg=msg, buffers=[])
    msg = dict(type='all_reprs_info', data=REPR_DICT)
    view._handle_custom_msg(msg=msg, buffers=[])
    msg = dict(type='stage_parameters', data=dict())
    view._handle_custom_msg(msg=msg, buffers=[])
    # test negative frame (it will be set to self.count - 1)
    view.frame = -1
    msg = dict(type='request_frame', data=dict())
    # async_message
    msg  = {'type': 'async_message', 'data': 'ok'}
    view._handle_custom_msg(msg, [])
    # render_image
    r = view.render_image()
    msg = {'type': 'image_data', 'ID': r.model_id, 'data': b'YmxhIGJsYQ=='}
    view._handle_custom_msg(msg, [])
    view.loaded = True
    view.show_only([
        0,
    ])
    view._js_console()
    view._get_full_params()

    # iter
    for c in view:
        assert isinstance(c, nv.widget.ComponentViewer)


@unittest.skipUnless(has_pytraj, 'skip if not having pytraj')
@unittest.skipUnless(has_mdtraj, 'skip if not having mdtraj')
def test_add_trajectory():
    view = nv.NGLWidget(default=False)

    def update_coords(view=view):
        view.frame = 1000
        view.frame = 0

    p_traj = pt.load(nv.datafiles.TRR, nv.datafiles.PDB)
    view.add_trajectory(p_traj)
    m_traj = md.load(nv.datafiles.XTC, top=nv.datafiles.PDB)
    view.add_trajectory(m_traj)
    # trigger updating coordinates
    update_coords()
    assert len(view._coordinates_dict.keys()) == 2
    if has_MDAnalysis:
        from MDAnalysis import Universe
        mda_traj = Universe(nv.datafiles.PDB, nv.datafiles.TRR)
        view.add_trajectory(mda_traj)
        update_coords()
        assert len(view._coordinates_dict.keys()) == 3
    if has_HTMD:
        from htmd import Molecule
        htmd_traj = Molecule(nv.datafiles.PDB)
        htmd_traj.filter('protein')
        view.add_trajectory(htmd_traj)
        update_coords()
        if has_MDAnalysis:
            assert len(view._coordinates_dict.keys()) == 4
        else:
            assert len(view._coordinates_dict.keys()) == 3


def test_API_promise_to_have_add_more_backend():
    @nv.register_backend('dummy')
    class MyLovelyClass(nv.Structure, nv.Trajectory):
        pass

    assert 'dummy' in nv.BACKENDS


def test_handling_n_components_changed():
    view = nv.NGLWidget()
    n_traj = get_simple_traj()
    view.add_trajectory(n_traj)
    # fake updating n_components and _repr_dict from front-end
    view._ngl_repr_dict = REPR_DICT
    view.n_components = 1
    view.player.widget_repr = view.player._make_widget_repr()
    view.remove_component(n_traj.id)
    # fake updating n_components from front-end
    view._ngl_repr_dict = {'c0': {}}
    view.n_components = 0


def test_base_adaptor():
    # abstract base class
    def func_0():
        nv.Structure().get_structure_string()

    def func_1():
        nv.Trajectory().get_coordinates(1)

    def func_2():
        nv.Trajectory().n_frames

    pytest.raises(NotImplementedError, func_0)
    pytest.raises(NotImplementedError, func_1)
    pytest.raises(NotImplementedError, func_2)


def test_coordinates_dict():
    traj = get_simple_traj()
    view = nv.NGLWidget(traj)
    view.frame = 1
    coords = view._coordinates_dict[0]
    aa_eq(coords, traj.get_coordinates(1))

    # dummy
    view._send_binary = False
    view._coordinates_dict = {0: coords}
    # increase coverage for IndexError: make index=1000 (which is larger than n_frames)
    view.player.interpolate = True
    view._set_coordinates(1000)


def test_load_data():
    view = default_view()

    # load blob with ext
    blob = open(nv.datafiles.PDB).read()
    view._load_data(blob, ext='pdb')

    # raise if passing blob but does not provide ext
    with pytest.raises(ValueError):
        view._load_data(blob)

    # raise if passing dummy name
    with pytest.raises(NameError):
        view._load_data(hahahaha)

    # load PyTrajectory
    t0 = get_simple_traj()
    view._load_data(t0)

    # load current folder
    view._load_data(get_fn('tz2.pdb'))


def test_representations():
    view = default_view()
    view.representations = DEFAULT_REPR
    view.add_cartoon()
    representations_2 = DEFAULT_REPR[:]
    representations_2.append({'type': 'cartoon', 'params': {'sele': 'all'}})
    _assert_dict_list_equal(view.representations, representations_2)

    # accept dict too (to specify seperate reprs for different component
    def func():
        view.representations = {'0': MagicMock()}
    assert view._dry_run(func)['methodName'] == '_set_representation_from_repr_dict'

    # Representations
    # make fake params
    try:
        view._ngl_repr_dict = {'c0': {'0': {'parameters': {}}}}
    except (KeyError, TraitError):
        # in real application, we are not allowed to assign values
        pass

    view._ngl_repr_dict = REPR_DICT
    representation_widget = RepresentationControl(view, 0, 0)
    representation_widget
    representation_widget._on_parameters_changed(change=dict(new=dict()))


def test_representation_control():
    view = nv.demo()
    repr_control = view._display_repr()

    repr_control.name = 'surface'
    repr_control.name = 'cartoon'
    repr_control.repr_index = 1
    repr_control.component_index = 1


def test_add_repr_shortcut():
    view = default_view()
    assert isinstance(view, nv.NGLWidget), 'must be instance of NGLWidget'

    # add
    view.add_cartoon(color='residueindex')
    view.add_rope(color='red')

    # update
    view.update_cartoon(opacity=0.4)
    view.update_rope(coor='blue')

    # remove
    view.remove_cartoon()
    view.remove_rope()


def test_color_scheme():
    view = nv.demo()
    scheme = nv.color._ColorScheme([['red', '1-6'], ['yellow', '20-30']],
                                   'what')
    view.clear()
    view.add_cartoon(color=scheme)


def test_add_new_shape():
    view = nv.NGLWidget()
    sphere = ('sphere', [0, 0, 9], [1, 0, 0], 1.5)
    arrow = ('arrow', [1, 2, 7], [30, 3, 3], [1, 0, 1], 1.0)
    c0 = view._add_shape([sphere, arrow], name='my_shape')

    # Shape
    c1 = view.shape.add_arrow([1, 2, 7], [30, 3, 3], [1, 0, 1], 1.0)
    assert len(view._ngl_component_ids) == 2
    view.remove_component(c0)
    assert len(view._ngl_component_ids) == 1

    view.remove_component(c1)
    assert len(view._ngl_component_ids) == 0


def test_add_buffer():
    view = nv.NGLWidget()
    view
    kwargs = {
        "position": [0, 0, 0, 1, 1, 1],
        "color": [1, 0, 0, 255, 0, 0],
        "radius": [1., 2.]
    }

    view.shape.add_buffer('sphere', **kwargs)


def test_remote_call():
    # how to test JS?
    view = default_view()
    view._remote_call('centerView', target='stage')

    fn = 'notebooks/tz2.pdb'
    kwargs = {'defaultRepresentation': True}
    view._remote_call('loadFile', target='stage', args=[
        fn,
    ], kwargs=kwargs)


def test_download_image():
    """just make sure it can be called
    """
    view = default_view()
    view.download_image('myname.png', 2, False, False, True)


def test_show_structure_file():
    view = nv.show_structure_file(nv.datafiles.PDB)


def test_show_file():
    view = nv.show_file(nv.datafiles.PDB)


def test_show_text():
    text = open(nv.datafiles.PDB).read()
    nv.show_text(text)


@unittest.skipUnless(has_ase, 'skip if not having ase')
def test_show_ase():
    from ase import Atom, Atoms
    dimer = Atoms([Atom('X', (0, 0, 0)), Atom('X', (0, 0, 1))])
    dimer.set_positions([(1, 2, 3), (4, 5, 6.2)])
    nv.show_ase(dimer)


@unittest.skipUnless(has_pymatgen, 'skip if not having pymatgen')
def test_show_pymatgen():
    from pymatgen.core import Lattice, Structure
    lattice = Lattice.cubic(4.2)
    structure = Structure(lattice, ["Cs", "Cl"],
                             [[0, 0, 0], [0.5, 0.5, 0.5]])
    view = nv.show_pymatgen(structure)
    view


@unittest.skipUnless(has_qcelemental, 'skip if not having qcelemental')
def test_show_qcelemental():
    import qcelemental as qcel

    mol = qcel.models.Molecule.from_data("He 0 0 0") 
    view = nv.show_qcelemental(mol)
    view


@unittest.skipUnless(has_bio, 'skip if not having biopython')
def test_show_biopython():
    from Bio.PDB import PDBParser
    parser = PDBParser()
    structure = parser.get_structure('protein', nv.datafiles.PDB)
    nv.show_biopython(structure)


@unittest.skipUnless(has_simpletraj, 'skip if not having simpletraj')
def test_show_simpletraj():
    traj = nv.SimpletrajTrajectory(nv.datafiles.XTC, nv.datafiles.GRO)
    view = nv.show_simpletraj(traj)
    view
    view.frame = 3


@unittest.skipUnless(has_mdtraj, 'skip if not having mdtraj')
def test_show_mdtraj():
    import mdtraj as md
    traj = md.load(nv.datafiles.PDB)
    view = nv.show_mdtraj(traj)


@unittest.skipUnless(has_HTMD, 'skip if not having HTMD')
def test_show_htmd():
    from htmd import Molecule
    fn = nv.datafiles.PDB
    traj = Molecule(fn)
    view = nv.show_htmd(traj)
    # trigger updating cooridnates
    view.frame = 100
    index = 0
    view.frame = index
    xyz_htmd = np.squeeze(traj.coords[:, :, index])
    aa_eq(view._coordinates_dict[0], xyz_htmd)


@unittest.skipUnless(has_MDAnalysis, 'skip if not having MDAnalysis')
def test_show_MDAnalysis():
    from MDAnalysis import Universe
    tn, fn = nv.datafiles.PDB, nv.datafiles.PDB
    u = Universe(fn, tn)
    view = nv.show_mdanalysis(u)


@unittest.skipUnless(has_parmed, 'skip if not having ParmEd')
def test_show_parmed():
    import parmed as pmd
    fn = nv.datafiles.PDB
    parm = pmd.load_file(fn)
    view = nv.show_parmed(parm)

    ngl_traj = nv.ParmEdTrajectory(parm)
    ngl_traj.only_save_1st_model = False
    ngl_traj.get_structure_string()


def test_encode_and_decode():
    xyz = np.arange(100).astype('f4')
    shape = xyz.shape

    b64_str = encode_base64(xyz)
    new_xyz = decode_base64(b64_str, dtype='f4', shape=shape)
    aa_eq(xyz, new_xyz)


def test_structure_file():
    for fn in [get_fn('tz2.pdb'), nv.datafiles.GRO]:
        content = open(fn, 'r').read()
        fs1 = nv.FileStructure(fn)
        assert content == fs1.get_structure_string()

    # gz
    fn = get_fn('tz2_2.pdb.gz')
    fs2 = nv.FileStructure(fn)
    content = gzip.open(fn).read()
    assert content == fs2.get_structure_string()


def test_camelize_parameters():
    view = nv.NGLWidget()
    view.parameters = dict(background_color='black')
    assert 'backgroundColor' in view._parameters


def test_component_for_duck_typing():
    # FIXME: deprecate duck typing?
    # syntax looks ugly.
    view = NGLWidget()
    traj = nv.SimpletrajTrajectory(nv.datafiles.XTC, nv.datafiles.GRO)

    # add 3 components (trajectory is a component)
    view.add_component(get_fn('tz2.pdb'))
    view.add_component(get_fn('tz2_2.pdb.gz'))
    view.add_trajectory(traj)
    view.component_0.add_representation('cartoon')

    c0 = view[0]
    c1 = view[1]
    assert hasattr(view, 'component_0')
    assert hasattr(view, 'component_1')
    assert hasattr(view, 'trajectory_0')
    assert hasattr(view.trajectory_0, 'n_frames')
    assert hasattr(view.trajectory_0, 'get_coordinates')
    assert hasattr(view.trajectory_0, 'get_structure_string')

    c0.show()
    c0.hide()

    # 2 components left
    view.remove_component(c0.id)
    # c1 become 1st component
    assert not hasattr(view, 'component_2')
    assert len(view._ngl_component_ids) == 2

    # negative indexing
    assert view[0]._index == c1._index


def test_trajectory_show_hide_sending_cooridnates():
    view = NGLWidget()

    traj0 = get_simple_traj()
    traj1 = get_simple_traj()

    view.add_trajectory(traj0)
    view.add_trajectory(traj1)

    for traj in view._trajlist:
        assert traj.shown

    view.frame = 1

    def copy_coordinate_dict(view):
        # make copy to avoid memory free
        return {k: v.copy() for k, v in view._coordinates_dict.items()}

    coordinates_dict = copy_coordinate_dict(view)
    aa_eq(coordinates_dict[0], traj0.get_coordinates(1))
    aa_eq(coordinates_dict[1], traj1.get_coordinates(1))

    # hide 0
    view.hide([
        0,
    ])
    assert not view._trajlist[0].shown
    assert view._trajlist[1].shown

    # update frame so view can update its coordinates
    view.frame = 2
    coordinates_dict = copy_coordinate_dict(view)
    assert coordinates_dict[0].shape[0] == 0
    aa_eq(coordinates_dict[1], traj1.get_coordinates(2))

    # hide 0, 1
    view.hide([0, 1])
    assert not view._trajlist[0].shown
    assert not view._trajlist[1].shown
    view.frame = 3
    coordinates_dict = copy_coordinate_dict(view)
    assert coordinates_dict[0].shape[0] == 0
    assert coordinates_dict[1].shape[0] == 0

    # slicing, show only component 1
    view[1].show()
    view.frame = 0
    assert not view._trajlist[0].shown
    assert view._trajlist[1].shown
    coordinates_dict = copy_coordinate_dict(view)
    assert coordinates_dict[0].shape[0] == 0
    aa_eq(coordinates_dict[1], traj1.get_coordinates(0))

    # show all
    view[1].show()
    view[0].show()
    view.show(indices='all')
    view.show(indices=[
        0,
    ])
    view.show(indices=[0, 1])
    view.frame = 1
    assert view._trajlist[1].shown
    coordinates_dict = copy_coordinate_dict(view)
    aa_eq(coordinates_dict[0], traj0.get_coordinates(1))
    aa_eq(coordinates_dict[1], traj1.get_coordinates(1))

    # hide all
    view[1].hide()
    view[0].hide()
    view.frame = 2
    assert not view._trajlist[0].shown
    assert not view._trajlist[1].shown
    coordinates_dict = copy_coordinate_dict(view)
    assert coordinates_dict[0].shape[0] == 0
    assert coordinates_dict[1].shape[0] == 0


def test_existing_js_files():
    from glob import glob
    jsfiles = glob(os.path.join(os.path.dirname(nv.__file__), 'static', '*js'))
    mapfiles = glob(
        os.path.join(os.path.dirname(nv.__file__), 'static', '*map'))

    assert len(jsfiles) == 2
    assert len(mapfiles) == 1


def test_add_structure():
    view = nv.NGLWidget()
    with pytest.raises(ValueError):
        # raise if not is instance of nv.Structure
        view.add_structure(nv.datafiles.PDB)


def test_add_struture_then_trajectory():
    view = nv.show_structure_file(get_fn('tz2.pdb'))
    view.loaded = True
    traj = get_simple_traj()
    view.add_trajectory(traj)
    view.frame = 3
    coords = view._coordinates_dict[1].copy()
    expected = traj.get_coordinates(3)
    aa_eq(coords, expected)
    view.loaded = False
    view.add_trajectory(traj)


def test_loaded_attribute():
    traj = get_simple_traj()
    structure = nv.FileStructure(nv.datafiles.PDB)

    # False, empty constructor
    view = nv.NGLWidget()
    view.loaded = False
    view.add_structure(structure)
    view.add_trajectory(traj)
    view

    # False, constructor with a single Structure
    view = nv.NGLWidget(structure)
    view.loaded = False
    view.add_trajectory(traj)
    view

    # True
    view = nv.NGLWidget()
    view.loaded = True
    view.add_structure(structure)
    view.add_trajectory(traj)
    view

    # False then True, empty constructor
    view = nv.NGLWidget()
    view.loaded = False
    view.add_structure(structure)
    view.loaded = True
    view.add_trajectory(traj)
    view

    # False then True, constructor with a Trajectory
    view = nv.NGLWidget(traj)
    view.loaded = False
    view.add_structure(structure)
    view.loaded = True
    view.add_trajectory(traj)
    view


def test_player_simple():
    view = default_view()

    # dummy
    component_slider = ipywidgets.IntSlider()
    repr_slider = ipywidgets.IntSlider()

    # dummy test
    player = nv.player.TrajectoryPlayer(view)
    player.smooth()
    player.camera = 'perspective'
    player.camera = 'orthographic'
    player.frame
    player.frame = 10
    player.parameters = dict(step=2)
    player._display()
    player._make_button_center()
    w = player._make_widget_preference()
    w.children[0].value = 1.
    player.widget_preference = None
    w = player._make_widget_preference()
    w.children[0].value = 1.
    player._show_download_image()
    player._make_text_picked()
    player._refresh(component_slider, repr_slider)
    player._make_widget_repr()
    player._make_resize_notebook_slider()
    player._make_button_export_image()
    player._make_repr_playground()
    player._make_widget_picked()
    player._make_export_image_widget()
    player._make_general_box()
    player._update_padding()
    player._real_time_update = True
    player._make_widget_repr()
    player.widget_component_slider
    player.widget_repr_slider
    player._create_all_tabs()
    player._create_all_widgets()
    player.widget_tab = None
    player._create_all_widgets()
    player._simplify_repr_control()

    player._real_time_update = True
    player.widget_repr_slider.value = 0
    player.widget_repr_slider.value = 1
    slider_notebook = player._make_resize_notebook_slider()
    slider_notebook.value = 300

    player.widget_repr_name.value = 'surface'
    player.widget_repr_name.value = 'cartoon'


def test_player_submit_text():
    """ test_player_click_button """
    view = nv.demo(gui=True)
    submit(view.player._make_command_box())


def test_player_click_button():
    """ test_player_click_button """
    view = nv.demo(gui=True)
    view
    view._ngl_repr_dict = REPR_DICT
    view.player._create_all_widgets()
    view.player.widget_export_image = view.player._make_button_export_image()
    button_iter = chain.from_iterable([
        view.player.widget_repr_control_buttons.children,
        [
            view.player._show_download_image(),
            view.player._make_button_center(),
            view.player.widget_export_image.children[0].children[0],
            view.player.widget_repr_add.children[0],
        ],
        [
            w for w in view.player.widget_preference.children
            if isinstance(w, Button)
        ],
    ])
    for button in button_iter:
        click(button)


def test_player_link_to_ipywidgets():
    view = default_view()

    int_text = IntText(2)
    float_text = BoundedFloatText(40, min=10)
    HBox([int_text, float_text])
    link((int_text, 'value'), (view.player, 'step'))
    link((float_text, 'value'), (view.player, 'delay'))

    assert view.player.step == 2
    assert view.player.delay == 40

    float_text.value = 100
    assert view.player.delay == 100

    float_text.value = 0.00
    # we set min=10
    assert view.player.delay == 10


def test_player_interpolation():
    view = default_view()

    view.player.interpolate = True
    assert view.player.iparams.get('type') == 'linear'
    assert view.player.iparams.get('step') == 1


def test_player_picked():
    view = nv.demo()
    s = dict(x=3)
    view.player.widget_picked = view.player._make_text_picked()
    view.picked = s
    assert view.player.widget_picked.value == '{"x": 3}'


def test_widget_utils():
    box = HBox()
    i0 = IntText()
    i0._ngl_name = 'i0'
    i1 = IntText()
    i1._ngl_name = 'i1'
    box.children = [i0, i1]

    assert i0 is widget_utils.get_widget_by_name(box, 'i0')
    assert i1 is widget_utils.get_widget_by_name(box, 'i1')

    box.children = [i1, i0]
    assert i0 is widget_utils.get_widget_by_name(box, 'i0')
    assert i1 is widget_utils.get_widget_by_name(box, 'i1')

    assert widget_utils.get_widget_by_name(box, 'i100') is None
    assert widget_utils.get_widget_by_name(None, 'i100') is None


def test_adaptor_raise():
    with pytest.raises(ValueError):
        nv.FileStructure('hellotheredda.pdb')


def test_theme():
    from nglview import theme
    # FIXME: fill me


def test_player_click_tab():
    view = nv.demo()
    gui = view.player._display()
    assert isinstance(gui, ipywidgets.Tab)

    for i, child in enumerate(gui.children):
        try:
            gui.selected_index = i
            assert isinstance(child, ipywidgets.Box)
        except TraitError:
            pass


def test_interpolate():
    # dummy test
    traj = get_simple_traj()
    interpolate.linear(0, 0.4, traj, step=1)


def dummy_test_to_increase_coverage():
    nv.__version__


def test_viewer_control():
    view = nv.demo()
    view

    mat = [11, 12, 13, 14, 21, 22, 23, 24, 31, 32, 33, 34, 41, 42, 43, 44]

    vector = [0, 1, 2]

    view.control.align(mat)
    view.control.rotate(mat)
    view.control.translate(vector)
    view.control.apply_matrix(mat)
    view.control.center(vector)
    view.control.orient(mat)
    view.control.zoom(0.3)
    view.control.rotate(mat)
    view.control.spin(vector, 0.1)


def test_queuing_messages():
    view = nv.NGLWidget()
    view.add_component(nv.datafiles.PDB)
    view.download_image()
    view
    assert [f._method_name for f in view._ngl_displayed_callbacks_before_loaded] == \
           [
            'loadFile',
            '_downloadImage']
    assert [f['methodName'] for f in view._ngl_msg_archive] == \
           ['loadFile']

    # display 2nd time
    view
    assert [f['methodName'] for f in view._ngl_msg_archive] == \
           ['loadFile']


@patch('nglview.NGLWidget._unset_serialization')
def test_write_html(mock_unset):
    from nglview.color import ColormakerRegistry as cm
    from nglview.theme import ThemeManager
    import ipywidgets.embed as embed

    tm = ThemeManager()
    traj0 = get_simple_traj()
    traj1 = get_simple_traj()
    view = nv.NGLWidget()
    view.add_trajectory(traj0)
    view.add_trajectory(traj1)
    view.gui_style = 'ngl'
    view._gui_theme = 'dark'
    display(view)
    fp = StringIO()

    with patch.object(embed, 'embed_snippet') as mock_embed:
        nv.write_html(fp, [view], frame_range=(0, 3))
        mock_embed.assert_called_with([tm, cm, view])
    mock_unset.assert_called_with()
    assert len(view._ngl_coordinate_resource[0]) == 3
    assert len(view._ngl_coordinate_resource[1]) == 3

    # box
    with patch.object(embed, 'embed_snippet') as mock_embed:
        nv.write_html(fp, [HBox([view])], frame_range=(0, 3))
        # FIXME: assertion?


def test_trim_messages():
    view = nv.demo()
    view.remove_component(view[0])
    assert view._ngl_msg_archive == []
    view.add_component(nv.datafiles.ALA3)
    assert len(view._ngl_msg_archive) == 1
    assert view._ngl_msg_archive[0]['methodName'] == 'loadFile'

    view = nv.demo()
    c = view.add_component(nv.datafiles.ALA3)
    view.remove_component(c)
    assert len(view._ngl_msg_archive) == 1
    assert view._ngl_msg_archive[0]['methodName'] == 'loadFile'


def test_fullscreen():
    v = nv.demo()
    fs = nv.widget.Fullscreen(v, [v])
    fs.fullscreen()
    with patch.object(v, 'handle_resize'):
        fs._fullscreen_changed(MagicMock(new=True, old=False))
        assert v.handle_resize.called
    # just run the code
    fs._fullscreen_changed(MagicMock(new=True, old=False))
    fs._fullscreen_changed(MagicMock(new=False, old=True))



import nglview as nv


def test_component_viewer():
    view = nv.NGLWidget()
    c = view.add_component(nv.datafiles.PDB)
    c._call("setPosition", [1, 2, 3])
    c.add_cartoon()
    view.remove_component(c)
    assert c._view is None



import os
from IPython.display import display
from ipywidgets import HTML, DOMWidget
from traitlets import Unicode, List, observe
from pathlib import Path

from ..base import _singleton, BaseWidget
from .._frontend import __frontend_version__


def _get_css_content(css_file):
    p = Path(__file__).resolve().parent / css_file
    return p.read_text()


@_singleton
class ThemeManager(BaseWidget):
    """EXPERIMENT
    """
    _view_name = Unicode("ThemeManagerView").tag(sync=True)
    _view_module = Unicode("nglview-js-widgets").tag(sync=True)
    _view_module_version = Unicode(__frontend_version__).tag(sync=True)
    _model_name = Unicode("ThemeManagerModel").tag(sync=True)
    _model_module = Unicode("nglview-js-widgets").tag(sync=True)
    _model_module_version = Unicode(__frontend_version__).tag(sync=True)
    _theme_css = Unicode(allow_none=True).tag(sync=True)
    # NOTE: Use `_theme_css` trait to sync backend and frontend to trigger them changed.
    # Use "sync" to make sure theme is created before displaying the NGLWidget.

    # overwrite BaseWidget's trait to avoid caling base method in frontend
    _msg_q = List().tag(sync=True)

    def __init__(self):
        super().__init__()
        self._theme = None
        display(self)

    @observe("_theme_css")
    def _on_theme_changed(self, _):
        """
        Without this method, we can't change the theme with below code

        # cell 1
        >>> import nglview as nv
        >>> 
        >>> view = nv.demo()
        >>> view.gui_style = 'ngl'
        >>> view._widget_theme.light()
        >>> # view._widget_theme.dark()
        >>> view

        # cell 2
        >>> import nglview as nv
        >>> 
        >>> view = nv.demo()
        >>> view.gui_style = 'ngl'
        >>> # view._widget_theme.light()
        >>> view._widget_theme.dark()
        >>> view
        """
        self._call('handleThemeChanged')

    def _ipython_display_(self, **kwargs):
        if self._ready:
            return
        else:
            super()._ipython_display_(**kwargs)

    def remove(self):
        self._js("""
            var ele = document.getElementById('nglview_style')
            if (ele){
                document.head.removeChild(ele)
            }
        """)
        self._theme = None
        self._theme_css = ''

    def dark(self):
        self._theme_css = _get_css_content('dark.css') + _get_css_content('main.css')
        self._theme = 'dark'

    def light(self):
        self._theme_css = _get_css_content('light.css') + _get_css_content('main.css')
        self._theme = 'light'



import os
from contextlib import contextmanager
import warnings
import os.path
import uuid
from functools import partial
from io import StringIO
from tempfile import mkstemp
from urllib.request import urlopen

import numpy as np

from . import config
from .base_adaptor import Structure, Trajectory
from .utils.py_utils import FileManager, tempfolder

__all__ = [
    'FileStructure',
    'TextStructure',
    'RdkitStructure',
    'PdbIdStructure',
    'ASEStructure',
    'BiopythonStructure',
    'IOTBXStructure',
    'IODataStructure',
    'QCElementalStructure',
    'Psi4Structure',
    'OpenbabelStructure',
    'RosettaStructure',
    'ProdyStructure',
    'SimpletrajTrajectory',
    'ProdyTrajectory',
    'MDTrajTrajectory',
    'PyTrajTrajectory',
    'ParmEdTrajectory',
    'MDAnalysisTrajectory',
    'HTMDTrajectory',
    'ASETrajectory',
    'SchrodingerStructure',
    'SchrodingerTrajectory',
    'register_backend',
]


@contextmanager
def mkstemp_wrapper(*args, **kwargs):
    # NamedTemporaryFile cannot be used here because it makes it impossible
    # on Windows to the file for writing. mkstemp is a bit less restrictive
    # in this regard.
    fd, fname = mkstemp(*args, **kwargs)
    yield fname
    # On windows, the file must be closed before it can be removed.
    os.close(fd)
    os.remove(fname)


def _get_structure_string(write_method, suffix='.pdb'):
    with mkstemp_wrapper(suffix=suffix) as fname:
        write_method(fname)
        with open(fname) as fh:
            return fh.read()


class register_backend:
    def __init__(self, package_name):
        # package_name must match exactly to your Python package
        self.package_name = package_name

    def __call__(self, cls):
        config.BACKENDS[self.package_name] = cls
        return cls


class FileStructure(Structure):
    def __init__(self, path):
        super().__init__()
        self.fm = FileManager(path)
        self.ext = self.fm.ext
        self.params = {}
        self.path = path

    def get_structure_string(self):
        return self.fm.read(force_buffer=True)


class TextStructure(Structure):
    def __init__(self, text, ext='pdb', params={}):
        super().__init__()
        self.path = ''
        self.ext = ext
        self.params = params
        self._text = text

    def get_structure_string(self):
        return self._text


@register_backend('rdkit')
class RdkitStructure(Structure):
    def __init__(self, rdkit_mol, ext="pdb", conf_id=-1):
        super().__init__()
        self.path = ''
        self.ext = ext
        self._conf_id = conf_id
        self.params = {}
        self._rdkit_mol = rdkit_mol

    def get_structure_string(self):
        from rdkit import Chem

        reader = (self.ext == "pdb") and \
                 Chem.MolToPDBBlock or Chem.MolToMolBlock
        return reader(self._rdkit_mol, confId=self._conf_id)


class PdbIdStructure(Structure):
    def __init__(self, pdbid):
        super().__init__()
        self.pdbid = pdbid
        self.ext = "cif"
        self.params = {}

    def get_structure_string(self):
        url = "http://www.rcsb.org/pdb/files/" + self.pdbid + ".cif"
        return urlopen(url).read().decode('utf-8')


class ASEStructure(Structure):
    def __init__(self, ase_atoms, ext='pdb', params={}):
        super().__init__()
        self.path = ''
        self.ext = ext
        self.params = params
        self._ase_atoms = ase_atoms

    def get_structure_string(self):
        return _get_structure_string(self._ase_atoms.write)


class IODataStructure(Structure):
    def __init__(self, obj):
        super().__init__()
        self._obj = obj

    def get_structure_string(self):
        """Require `ase` package
        """
        import ase.io
        with mkstemp_wrapper(suffix='.xyz') as fname:
            self._obj.to_file(fname)
            return ASEStructure(ase.io.read(fname)).get_structure_string()


class QCElementalStructure(Structure):
    def __init__(self, obj):
        super().__init__()
        self._obj = obj
        self.ext = 'sdf'

    def get_structure_string(self):
        return self._obj.to_string('nglview-sdf')


class Psi4Structure(QCElementalStructure):
    pass


class OpenbabelStructure(Structure):
    def __init__(self, obj):
        super().__init__()
        self._obj = obj

    def get_structure_string(self):
        try:
            # Open Babel >= 3.0.0
            from openbabel import openbabel
        except ImportError:
            import openbabel
        oc = openbabel.OBConversion()
        oc.SetOutFormat('pdb')
        write = partial(oc.WriteFile, self._obj)
        return _get_structure_string(write)


class BiopythonStructure(Structure):
    def __init__(self, entity, ext='pdb', params={}):
        super().__init__()
        self.path = ''
        self.ext = ext
        self.params = params
        self._entity = entity

    def get_structure_string(self):
        from Bio.PDB import PDBIO
        from io import StringIO
        io_pdb = PDBIO()
        io_pdb.set_structure(self._entity)
        io_str = StringIO()
        io_pdb.save(io_str)
        return io_str.getvalue()


class IOTBXStructure(Structure):
    def __init__(self, obj, ext='pdb', params={}):
        """
        obj must have as_pdb_string method
        """
        super().__init__()
        self.path = ''
        self.ext = ext
        self.params = params
        self._mol = obj

    def get_structure_string(self):
        return self._mol.as_pdb_string()


class RosettaStructure(Structure):
    def __init__(self, pose, ext='pdb', params={}):
        # type: (pyrosetta.rosetta.core.pose.Pose, str, Dict) -> None
        super().__init__()
        self.path = ''
        self.ext = ext
        self.params = params
        self._mol = pose

    def get_structure_string(self):
        return _get_structure_string(self._mol.dump_pdb)


@register_backend('prody')
class ProdyStructure(Structure):
    def __init__(self, obj):
        super().__init__()
        self._obj = obj
        self.ext = 'pdb'
        self.params = {}

    def get_structure_string(self):
        import prody

        def write(fname):
            if isinstance(self._obj, prody.Ensemble):
                st = self._obj[0]
            else:
                st = self._obj
            prody.writePDB(fname, st)

        return _get_structure_string(write)


@register_backend('prody')
class ProdyTrajectory(Trajectory, ProdyStructure):
    def __init__(self, obj):
        ProdyStructure.__init__(self, obj)

    @property
    def n_frames(self):
        return self._obj.numConfs()

    def get_coordinates(self, index):
        return self._obj.getConformation(index).getCoords()


@register_backend('simpletraj')
class SimpletrajTrajectory(Trajectory, Structure):
    '''simpletraj adaptor.

    Examples
    --------
    >>> import nglview as nv
    >>> t = nv.SimpletrajTrajectory(nv.datafiles.XTC, nv.datafiles.GRO)
    >>> w = nv.NGLWidget(t)
    >>> w
    '''

    def __init__(self, path, structure_path):
        try:
            import simpletraj
        except ImportError as e:
            raise ImportError(
                "'SimpletrajTrajectory' requires the 'simpletraj' package")
        self.traj_cache = simpletraj.trajectory.TrajectoryCache()
        self.path = path
        self._structure_path = structure_path
        self.ext = os.path.splitext(structure_path)[1][1:]
        self.params = {}
        self.trajectory = None
        self.id = str(uuid.uuid4())
        try:
            self.traj_cache.get(os.path.abspath(self.path))
        except Exception as e:
            raise e

    def get_coordinates(self, index):
        traj = self.traj_cache.get(os.path.abspath(self.path))
        frame = traj.get_frame(index)
        return frame["coords"]

    def get_structure_string(self):
        return open(self._structure_path).read()

    @property
    def n_frames(self):
        traj = self.traj_cache.get(os.path.abspath(self.path))
        return traj.numframes


@register_backend('mdtraj')
class MDTrajTrajectory(Trajectory, Structure):
    '''mdtraj adaptor.

    Example
    -------
    >>> import nglview as nv
    >>> import mdtraj as md
    >>> traj = md.load(nv.datafiles.XTC, nv.datafiles.GRO)
    >>> t = MDTrajTrajectory(traj)
    >>> w = nv.NGLWidget(t)
    >>> w
    '''

    def __init__(self, trajectory):
        self.trajectory = trajectory
        self.ext = "pdb"
        self.params = {}
        self.id = str(uuid.uuid4())

    def get_coordinates(self, index):
        return 10 * self.trajectory.xyz[index]

    @property
    def n_frames(self):
        return self.trajectory.n_frames

    def get_structure_string(self):
        return _get_structure_string(self.trajectory[0].save_pdb)


@register_backend('pytraj')
class PyTrajTrajectory(Trajectory, Structure):
    '''PyTraj adaptor.

    Example
    -------
    >>> import nglview as nv
    >>> import pytraj as pt
    >>> traj = pt.load(nv.datafiles.TRR, nv.datafiles.PDB)
    >>> t = nv.PyTrajTrajectory(traj)
    >>> w = nv.NGLWidget(t)
    >>> w
    '''

    def __init__(self, trajectory):
        self.trajectory = trajectory
        self.ext = "pdb"
        self.params = {}
        self.id = str(uuid.uuid4())

    def get_coordinates(self, index):
        return self.trajectory[index].xyz

    @property
    def n_frames(self):
        return self.trajectory.n_frames

    def get_structure_string(self, index=0):
        return _get_structure_string(self.trajectory[index:index + 1].save)


@register_backend('parmed')
class ParmEdStructure(Structure):
    def __init__(self, structure):
        self._structure = structure
        self.only_save_1st_model = True

    def get_structure_string(self):
        # only write 1st model
        with mkstemp_wrapper(suffix='.pdb') as fname:
            if self.only_save_1st_model:
                self._structure.write_pdb(
                    fname, coordinates=self._structure.coordinates)
            else:
                self._structure.write_pdb(fname)
            with open(fname) as fh:
                return fh.read()


@register_backend('parmed')
class ParmEdTrajectory(Trajectory, ParmEdStructure):
    '''ParmEd adaptor.
    '''

    def __init__(self, trajectory):
        ParmEdStructure.__init__(self, trajectory)
        self.trajectory = self._structure = trajectory
        self.ext = "pdb"
        self.params = {}
        # only call get_coordinates once
        self._xyz = trajectory.get_coordinates()
        self.id = str(uuid.uuid4())

    def get_coordinates(self, index):
        return self._xyz[index]

    @property
    def n_frames(self):
        return len(self._xyz)


@register_backend('MDAnalysis')
class MDAnalysisTrajectory(Trajectory, Structure):
    '''MDAnalysis adaptor.

    Can take either a Universe or AtomGroup.

    Example
    -------
    >>> import nglview as nv
    >>> import MDAnalysis as mda
    >>> u = mda.Universe(nv.datafiles.GRO, nv.datafiles.XTC)
    >>> prot = u.select_atoms('protein')
    >>> t = nv.MDAnalysisTrajectory(prot)
    >>> w = nv.NGLWidget(t)
    >>> w
    '''

    def __init__(self, atomgroup):
        self.atomgroup = atomgroup
        self.ext = "pdb"
        self.params = {}
        self.id = str(uuid.uuid4())

    def get_coordinates(self, index):
        self.atomgroup.universe.trajectory[index]
        xyz = self.atomgroup.atoms.positions
        return xyz

    @property
    def n_frames(self):
        return self.atomgroup.universe.trajectory.n_frames

    def get_structure_string(self):
        try:
            import MDAnalysis as mda
        except ImportError:
            raise ImportError(
                "'MDAnalysisTrajectory' requires the 'MDAnalysis' package")
        u = self.atomgroup.universe
        u.trajectory[0]
        f = mda.lib.util.NamedStream(StringIO(), 'tmp.pdb')
        atoms = self.atomgroup.atoms
        # add PDB output to the named stream
        with mda.Writer(f, atoms.n_atoms, multiframe=False) as w:
            with warnings.catch_warnings():
                warnings.simplefilter('ignore')
                w.write(atoms)
        # extract from the stream
        pdb_string = f.read()
        return pdb_string


@register_backend('htmd')
class HTMDTrajectory(Trajectory):
    '''HTMD adaptor.

    Takes a Molecule object.

    Example
    -------
    >>> import nglview as nv
    >>> from htmd import Molecule
    >>> mol = Molecule(nv.datafiles.PDB)
    >>> t = nv.HTMDTrajectory(mol)
    >>> w = nv.NGLWidget(t)
    >>> w
    '''

    def __init__(self, mol):
        self.mol = mol
        self.ext = "pdb"
        self.params = {}
        self.id = str(uuid.uuid4())

    def get_coordinates(self, index):
        return np.squeeze(self.mol.coords[:, :, index])

    @property
    def n_frames(self):
        return self.mol.numFrames

    def get_structure_string(self):
        return _get_structure_string(self.mol.write)


@register_backend('ase')
class ASETrajectory(Trajectory, Structure):
    '''asetraj adaptor.

    Examples
    --------
    >>> import nglview as nv
    >>> from ase.io.trajectory import Trajectory
    >>> traj = Trajectory(nv.datafiles.ASE_Traj)
    >>> t = nv.ASETrajectory(traj)
    >>> w = nv.NGLWidget(t)
    >>> w.add_spacefill()
    >>> w
    '''

    def __init__(self, ase_traj):
        self.ext = 'pdb'
        self.params = {}
        self.trajectory = ase_traj
        self.id = str(uuid.uuid4())

    def get_coordinates(self, index):
        return self.trajectory[index].positions

    def get_structure_string(self):
        return _get_structure_string(self.trajectory[0].write)

    @property
    def n_frames(self):
        return len(self.trajectory)


@register_backend('schrodinger')
class SchrodingerStructure(Structure):
    '''Only read first structure
    '''

    def __init__(self, structure, ext="pdb"):
        super().__init__()
        self.path = ''
        self.ext = ext
        self.params = {}
        self._schrodinger_structure = structure

    def get_structure_string(self):
        # NOTE: For some reasons, _get_structure_string always return
        # empty string in this case.
        with tempfolder():
            pdb_fn = 'tmp.pdb'
            self._schrodinger_structure.write(pdb_fn)
            with open(pdb_fn) as fh:
                content = fh.read()
        return content

        return _get_structure_string(self._schrodinger_structure.write)


@register_backend('schrodinger')
class SchrodingerTrajectory(SchrodingerStructure, Trajectory):
    """Require `parmed` package.
    """

    def __init__(self, structure, traj):
        super().__init__(structure)
        self._traj = traj

    @property
    def n_frames(self):
        return len(self._traj) if self._traj else 1

    def get_coordinates(self, index):
        return self._traj[index].pos()

    def get_structure_string(self):
        """Require `parmed` package.
        """
        import parmed as pmd
        c = self._schrodinger_structure
        s = pmd.Structure()
        fsys = c.fsys_ct if hasattr(c, 'fsys_ct') else c
        for atom in fsys.atom:
            parm_atom = pmd.Atom(name=atom.pdbname.strip(),
                                 atomic_number=atom.atomic_number)
            s.add_atom(parm_atom,
                       atom.pdbres.strip(),
                       atom.resnum,
                       chain=atom.chain)
        s.coordinates = fsys.getXYZ()
        return ParmEdStructure(s).get_structure_string()

    @classmethod
    def from_files(cls, cms_fname, traj_fname):
        from schrodinger.application.desmond.packages import topo, traj
        _, model = topo.read_cms(cms_fname)
        traj = traj.read_traj(traj_fname)
        return cls(model, traj)
